# -*- coding: utf-8 -*-
"""
@author:    Panos Athanasiou
Date:       04/07/2022
"""

import datetime
import json
import os
import time

import numpy as np
import pandas as pd
import requests
from pyproj import CRS

from urllib.request import urlretrieve

# from cht_meteo import gfs_forecast_0p25

class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = []
        self.x = None
        self.y = None
        self.crs = None
        self.val = None
        self.src = []


def download(param_list, lon_range, lat_range, time_range, cycle_time):
    """Function to download coamps-tc forecasts using the scrubber tool functions
    Right now resolution is hardcoded to 0.25 degrees but this can change"""

    base_url = "https://coamps-tc-data.s3.us-east-2.amazonaws.com/index.html#deterministic/realtime"
    base_url = base_url + "#deterministic/realtime"

    # Get year from cycle_time
    year = cycle_time.year
    cycle_time = cycle_time.strftime("%Y%m%d%H")
    storm_number = "02L"

    tar_file = storm_number + "_" + cycle_time + "_netcdf.tar"

    url = base_url + "/" + str(year) + "/" + storm_number + "/" + cycle_time + "/" + tar_file

    # Download the tar file from url to a temporary directory
    urlretrieve(url, tar_file)


    





#     fill_values = dict(
#         wind=0.0, precipitation=0.0, barometric_pressure=102000.0
#     )  # fill values to use in case of nan
#     units = dict(
#         wind="m/s", precipitation="kg.m-2.hour-1", barometric_pressure="Pa"
#     )  # units

#     apikey = "GrPBTsclg71qgDvw4ojGf4mTwZmkbqNu6jcq7T9V"  # Deltares apikey hardcoded for the time being
#     endpoint = "https://api.metget.zachcobell.com"

#     # Connect parameters names with coamps-tc names
#     param_names = {
#         "wind": "wind_pressure",
#         "barometric_pressure": "wind_pressure",
#         "precipitation": "rain",
#     }
#     variables = list(
#         np.unique([param_names[name] for name in param_list])
#     )  # ['wind_pressure', 'rain']

#     requested_times = (
#         pd.date_range(start=time_range[0], end=time_range[1], freq="3H")
#         .to_pydatetime()
#         .tolist()
#     )

#     timestep = int(
#         (requested_times[1] - requested_times[0]).total_seconds()
#     )  # get time-step of requested time assuming that the time-step stays the same
#     ntime = len(requested_times)

#     # Get storms from endpoint metadata
#     storms = check_coamps(apikey, endpoint)

#     # Check if forecast exists for the selected dates and make a dataframe
#     fr = pd.DataFrame(index=requested_times)
#     fr["storm_id"] = None

#     for i, ti in enumerate(requested_times):
#         fr.loc[ti, "storm_id"] = []
#         for j, storm_id in enumerate(storms):
#             # The MetGet tool seems to have changed ...
#             storm = storms[storm_id]
#             ####### CHECK WHICH times to use for range ####################
#             t1 = datetime.datetime.strptime(
#                 storm["min_forecast_date"], "%Y-%m-%d %H:%M:%S"
#             )
#             t2 = datetime.datetime.strptime(
#                 storm["max_forecast_date"], "%Y-%m-%d %H:%M:%S"
#             )
#             # t1 = datetime.datetime.strptime(storm['latest_complete_forecast_start'], "%Y-%m-%d %H:%M:%S")
#             # t2 = datetime.datetime.strptime(storm['latest_complete_forecast_end'], "%Y-%m-%d %H:%M:%S")
#             belongs = (
#                 t1.replace(tzinfo=datetime.timezone.utc)
#                 <= ti.replace(tzinfo=datetime.timezone.utc)
#                 <= t2.replace(tzinfo=datetime.timezone.utc)
#             )
#             if belongs:
#                 fr.loc[ti, "storm_id"].append(storm_id)

#     # Check if there is a txt file with the priority storm
#     file_storm = r"p:\11206085-onr-fhics\03_cosmos\Priority_Storm.txt"  # !!! Hard-coded
#     if os.path.exists(file_storm):
#         with open(file_storm, "r") as f:
#             lines = f.readlines()
#         f.close()
#         pr_st_txt = lines[0]
#     else:
#         pr_st_txt = ""
#     # Get storm priority from noaa
#     pr_st_noaa = get_priority_storm()
#     method = "first"
#     for i, ti in enumerate(requested_times):
#         if not fr.loc[
#             ti, "storm_id"
#         ]:  # if there is not coamps storm make None so gfs is used
#             fr.loc[ti, "storm_id"] = None
#         else:
#             if (
#                 pr_st_txt in fr.loc[ti, "storm_id"]
#             ):  # if the priority storm is available that timestep use it
#                 fr.loc[ti, "storm_id"] = pr_st_txt
#                 continue
#             if (
#                 pr_st_noaa in fr.loc[ti, "storm_id"]
#             ):  # if the priority storm is available that timestep use it
#                 fr.loc[ti, "storm_id"] = pr_st_noaa
#                 continue
#             # else used method defined
#             if method == "first":
#                 fr.loc[ti, "storm_id"] = fr.loc[ti, "storm_id"][0]
#             elif method == "last":
#                 fr.loc[ti, "storm_id"] = fr.loc[ti, "storm_id"][-1]

#     storm_log = np.array(
#         [v is not None for v in fr["storm_id"].to_numpy()]
#     )  # check for which times there are forecasts from coamps-tc
#     if all(~storm_log):  # If no coamps-tc data are available, gfs data are downloaded
#         datasets = gfs_forecast_0p25.download(
#             param_list, lon_range, lat_range, time_range, cycle_time
#         )
#         print(
#             "Could not find any coamps-tc data in requested time range! Using gfs instead"
#         )
#         return datasets

#     # Else if there are coamps-tc data see which storms (models) to use
#     storms_id = np.unique(fr["storm_id"][storm_log])
#     dss = {}  # prepare dictionary to save the datasets

#     # Loop for the case there is more than one storm in requested times and get storm forecast

#     for i, st in enumerate(storms_id):
#         dss[st] = {}
#         #        storm = storms[[storm_i for storm_i in range(len(storms)) if storms[storm_i]['storm'] == st][0]]
#         storm = storms[st]
#         t1 = datetime.datetime.strptime(
#             storm["min_forecast_date"], "%Y-%m-%d %H:%M:%S"
#         ).replace(tzinfo=datetime.timezone.utc)
#         t2 = datetime.datetime.strptime(
#             storm["max_forecast_date"], "%Y-%m-%d %H:%M:%S"
#         ).replace(tzinfo=datetime.timezone.utc)
#         # This parts ensures that requested data are within the model range
#         t1 = max(requested_times[0], t1)
#         t2 = min(requested_times[-1], t2)
#         if t1 == t2:  # scrubber tool cannot handle single time
#             t1 = t1 - datetime.timedelta(hours=3)
#         for ii, var in enumerate(variables):
#             ds = met_get(
#                 domain=[
#                     [
#                         "coamps-{}".format(st),
#                         resolution,
#                         lon_range[0],
#                         lat_range[0],
#                         lon_range[1],
#                         lat_range[1],
#                     ]
#                 ],  # list is needed here!
#                 start=t1.strftime("%Y-%m-%d %H:%M:%S"),
#                 end=t2.strftime("%Y-%m-%d %H:%M:%S"),
#                 timestep=timestep,
#                 variable=var,
#                 apikey=apikey,
#                 endpoint=endpoint,
#                 output="coamps_{}_{}".format(
#                     st, var
#                 ),  # this is used for the online file creation
#             )
#             if ds is not None:
#                 dss[st][var] = ds[0]
#             else:
#                 dss[st][var] = None

#     # if no data could be downloaded (CHECK AGAIN WHAT SHOULD HAPPEN IN THIS CASE)
#     if all(
#         [dss[k][kk] is None for k in list(dss.keys()) for kk in list(dss[k].keys())]
#     ):
#         datasets = gfs_forecast_0p25.download(
#             param_list, lon_range, lat_range, time_range, cycle_time
#         )
#         print(
#             "Could not find any coamps-tc data in requested time range! Using gfs instead"
#         )
#         return datasets
#     else:  # Else if there are available data get lon lat info
#         for key in dss[storms_id[0]].keys():
#             if dss[storms_id[0]][key] is not None:
#                 data0 = dss[storms_id[0]][key]
#                 lon = np.array(data0["lon"])
#                 lat = np.array(data0["lat"])
#                 if lat[1] - lat[0] > 0:  # lat should be in descending order
#                     lat = lat[::-1]
#                     reverse = True
#                 else:
#                     reverse = False
#                 lon = (lon + 360) % 360  # added this to be consistent with gfs
#                 nrows = len(lat)
#                 ncols = len(lon)

#     # Prepare datasets
#     datasets = []
#     for param in param_list:
#         dataset = Dataset()
#         dataset.crs = CRS.from_epsg(4326)
#         dataset.quantity = param
#         dataset.x = lon
#         dataset.y = lat
#         dataset.time = pd.to_datetime(
#             [tt.replace(tzinfo=None) for tt in requested_times]
#         ).to_pydatetime()
#         if dataset.quantity == "wind":
#             dataset.u = np.empty((ntime, nrows, ncols))
#             dataset.u[:] = np.NaN
#             dataset.v = np.empty((ntime, nrows, ncols))
#             dataset.v[:] = np.NaN
#         else:
#             dataset.val = np.empty((ntime, nrows, ncols))
#             dataset.val[:] = np.NaN
#         datasets.append(dataset)

#     for it, time_i in enumerate(requested_times):
#         storm_id = fr["storm_id"][it]

#         if storm_id is None:
#             try:
#                 #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
#                 datasets_gfs = gfs_forecast_0p25.download(
#                     param_list, lon_range, lat_range, [time_i, time_i], cycle_time
#                 )
#                 for ind, param in enumerate(param_list):
#                     if param == "wind":
#                         datasets[ind].u[it, :, :] = datasets_gfs[ind].u
#                         datasets[ind].v[it, :, :] = datasets_gfs[ind].v
#                     else:
#                         datasets[ind].val[it, :, :] = datasets_gfs[ind].val
#                     datasets[ind].unit = datasets_gfs[ind].unit
#                     datasets[ind].src.append("gfs")
#                 print(
#                     "No coamps-tc forecast available for {} --> using gfs forecast instead !!!".format(
#                         time_i
#                     )
#                 )

#             except Exception:
#                 print("gfs data not available")
#             continue

#         # Loop through requested parameters
#         for ind, param in enumerate(param_list):
#             dataset = datasets[ind]

#             try:
#                 okay = False
#                 makezeros = False
#                 #  First check if there are coamps-tc data available for this parameter
#                 if dss[storm_id][param_names[param]] is not None:
#                     okay = True
#                     model_t = pd.to_datetime(
#                         dss[storm_id][param_names[param]].time.values
#                     ).to_pydatetime()
#                     model_t_ind = np.where(model_t == time_i.replace(tzinfo=None))[0][0]
#                 else:
#                     makezeros = True

#                 if okay:
#                     dataset.coampsflag = True
#                     dataset.src.append("coamps_{}".format(storm_id))
#                     if param == "wind":
#                         data = dss[storm_id][param_names[param]]
#                         u = data["wind_u"]
#                         v = data["wind_v"]
#                         dataset.unit = u.units

#                         if np.any(np.isnan(u)) or np.any(
#                             np.isnan(v)
#                         ):  # check if there are nans and fill them up
#                             u = u.fillna(fill_values[param])
#                             v = v.fillna(fill_values[param])

#                         u = u.metpy.unit_array.squeeze()
#                         v = v.metpy.unit_array.squeeze()

#                         if reverse:
#                             dataset.u[it, :, :] = np.array(u[model_t_ind, ::-1, :])
#                             dataset.v[it, :, :] = np.array(v[model_t_ind, ::-1, :])
#                         else:
#                             dataset.u[it, :, :] = np.array(u[model_t_ind, :, :])
#                             dataset.v[it, :, :] = np.array(v[model_t_ind, :, :])
#                     else:
#                         if param == "barometric_pressure":
#                             var_name = "mslp"
#                         elif param == "precipitation":
#                             var_name = "rain"  # Check when available in scrubber tool!
#                         data = dss[storm_id][param_names[param]]
#                         val = data[var_name]

#                         # Added this check to ensure that pressure is in Pa
#                         if param == "barometric_pressure":
#                             if val.units == "mb":
#                                 val = val * 100
#                                 val.attrs["units"] = "Pa"
#                         dataset.unit = val.units

#                         if (
#                             param == "precipitation"
#                         ):  # Added this check to ensure that nan in precipitation are correctly interpreted
#                             val = val.where(val >= 0, np.nan)

#                         if np.any(
#                             np.isnan(val)
#                         ):  # check if there are nans and fill them up
#                             val = val.fillna(fill_values[param])

#                         val = np.array(val.metpy.unit_array.squeeze())

#                         if reverse:
#                             dataset.val[it, :, :] = np.array(val[model_t_ind, ::-1, :])
#                         else:
#                             dataset.val[it, :, :] = np.array(val[model_t_ind, :, :])

#                 elif makezeros:  # add zeros
#                     try:
#                         #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
#                         datasets_gfs = gfs_forecast_0p25.download(
#                             param_list,
#                             lon_range,
#                             lat_range,
#                             [time_i, time_i],
#                             cycle_time,
#                         )
#                         if param == "wind":
#                             dataset.u[it, :, :] = datasets_gfs[ind].u
#                             dataset.v[it, :, :] = datasets_gfs[ind].v
#                         else:
#                             dataset.val[it, :, :] = datasets_gfs[ind].val
#                         dataset.unit = datasets_gfs[ind].unit
#                         dataset.src.append("gfs")
#                         print(
#                             param
#                             + " was not found on server for {} --> using gfs forecast data instead".format(
#                                 time_i
#                             )
#                         )
#                     except Exception:
#                         if param == "wind":
#                             dataset.u[:] = fill_values[param]
#                             dataset.v[:] = fill_values[param]
#                             dataset.unit = units[param]
#                             dataset.src.append("default_value")
#                             print(
#                                 param
#                                 + " was not found on server for {} --> using {} {} instead !!!".format(
#                                     time_i, fill_values[param], units[param]
#                                 )
#                             )

#                         else:
#                             dataset.val[:] = fill_values[param]
#                             dataset.unit = units[param]
#                             dataset.src.append("default_value")
#                             print(
#                                 param
#                                 + " was not found on server ... --> using {} {} instead !!!".format(
#                                     fill_values[param], units[param]
#                                 )
#                             )

#             except Exception:
#                 print("Could not download data")

#     return datasets


# def get_priority_storm():
#     """Find the storm with the highest priority from NOAA"""
#     tcvitals = requests.get("https://ftp.nhc.noaa.gov/atcf/com/tcvitals").text
#     splits = [line.split() for line in tcvitals.split("\n")[:-1]]  # last line is empty
#     priority = pd.DataFrame(splits)
#     priority = priority.iloc[
#         [i for i in range(len(priority)) if "L" in priority.iloc[i, 1]], :
#     ]  # check only in atlantic basin storms
#     try:
#         pr_st_noaa = priority.iloc[0, 1]  # define the top L storm as the priority one
#         return pr_st_noaa
#     except Exception:
#         return None


# def check_coamps(apikey, endpoint):
#     """Read metadata of available forecasts from endpoint and returns a dictionary of the available storms"""
#     headers = {"x-api-key": apikey}
#     response_API = requests.get("{}/status".format(endpoint), headers=headers)
#     data = json.loads(response_API.text)
#     storms = data["body"]["data"]["metget"]["coamps-tc"]

#     return storms


# def parse_domain_data(domain_list: list, level) -> dict:
#     import warnings

#     AVAILABLE_MODELS = {
#         "gfs": "gfs-ncep",
#         "nam": "nam-ncep",
#         "hwrf": "hwrf",
#         "coamps": "coamps-tc",
#     }

#     model = domain_list[0]
#     if "hwrf" in model:
#         storm = model.split("-")[1]
#         model = "hwrf"
#         warnings.warn(
#             "HWRF not fully supported yet. Use at your own risk.", RuntimeWarning
#         )
#     elif "coamps" in model:
#         storm = model.split("-")[1]
#         model = "coamps"
#         # warnings.warn(
#         #     "COAMPS not fully supported yet. Use at your own risk.", RuntimeWarning
#         # )

#     res = float(domain_list[1])
#     x0 = float(domain_list[2])
#     y0 = float(domain_list[3])
#     x1 = float(domain_list[4])
#     y1 = float(domain_list[5])

#     if model not in AVAILABLE_MODELS.keys():
#         raise RuntimeError("Specified model '" + model + "' is not available")

#     xmax = max(x0, x1)
#     xmin = min(x0, x1)
#     ymax = max(y0, y1)
#     ymin = min(y0, y1)
#     res = abs(res)
#     if res <= 0:
#         raise RuntimeError("Specified model resolution is invalid")

#     if model == "hwrf" or model == "coamps":
#         return {
#             "name": AVAILABLE_MODELS[model] + "-" + storm,
#             "service": AVAILABLE_MODELS[model],
#             "storm": storm,
#             "x_init": xmin,
#             "y_init": ymin,
#             "x_end": xmax,
#             "y_end": ymax,
#             "di": res,
#             "dj": res,
#             "level": level,
#         }
#     else:
#         return {
#             "name": model,
#             "service": AVAILABLE_MODELS[model],
#             "x_init": xmin,
#             "y_init": ymin,
#             "x_end": xmax,
#             "y_end": ymax,
#             "di": res,
#             "dj": res,
#             "level": level,
#         }


# def make_metget_request(endpoint, apikey, request_json):
#     headers = {"x-api-key": apikey}
#     r = requests.post(endpoint + "/build", headers=headers, json=request_json)
#     if r.status_code != 200:
#         raise RuntimeError(
#             "Request to MetGet was returned status code = " + str(r.status_code)
#         )
#     return_data = json.loads(r.text)
#     data_id = return_data["body"]["request_id"]
#     status_code = return_data["statusCode"]
#     if status_code != 200:
#         with open("metget.debug", "a") as f:
#             f.write("[WARNING]: MetGet returned status code " + str(status_code) + "\n")
#             f.write(str(return_data["body"]["error_text"]))
#     return data_id, status_code


# def download_metget_data(data_id, endpoint, apikey, sleeptime, max_wait):
#     from datetime import datetime, timedelta

#     # from netCDF4 import Dataset
#     import xarray as xr

#     # ...Status check
#     headers = {"x-api-key": apikey}
#     request_json = {"request": data_id}

#     # ...Wait time
#     end_time = datetime.utcnow() + timedelta(hours=max_wait)

#     # ...Wait for request data to appear
#     tries = 0
#     data_ready = False
#     status = None
#     # print("Waiting for request id: ", data_id, flush=True)
#     while datetime.utcnow() <= end_time:
#         tries += 1
#         try:
#             # print(
#             #     "["
#             #     + datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")
#             #     + "]: Checking request status...(n="
#             #     + str(tries)
#             #     + "): ",
#             #     flush=True,
#             #     end="",
#             # )
#             response = requests.post(
#                 endpoint + "/check", headers=headers, json=request_json
#             )
#             json_response = json.loads(response.text)
#             status = json_response["body"]["status"]
#             data_url = json_response["body"]["destination"]
#             # print(status, flush=True)
#             if status == "completed":
#                 # ...Parse the return to get data
#                 data_ready = True
#                 flist_url = data_url + "/filelist.json"
#                 u = requests.get(flist_url)
#                 if u.status_code == 200:
#                     return_data = json.loads(u.text)
#                     # Turned this off to avoid saving this file
#                     # with open("filelist.json", "w") as jsonfile:
#                     #     jsonfile.write(
#                     #         json.dumps(return_data, indent=2, sort_keys=True)
#                     #     )
#                     break
#             elif status == "error":
#                 # print("Request could not be completed")
#                 return
#             else:
#                 time.sleep(sleeptime)
#                 continue
#         except KeyboardInterrupt:
#             # print("[ERROR]: Process was ended by the user")
#             raise

#     # ...Download files
#     if data_ready:
#         file_list = return_data["output_files"]
#         ds_list = []
#         for f in file_list:
#             # print("[{:s}]: Getting file: {:s}".format(time_stamp, f), flush=True)
#             url = data_url + "/" + f
#             # ds = Dataset('name', memory=requests.get(url).content)
#             try:
#                 ds = xr.open_dataset(
#                     url + "#mode=bytes"
#                 )  # Added this last part to allow opening with xarray
#             except Exception:
#                 from netCDF4 import Dataset

#                 data = requests.get(url).content
#                 ds0 = Dataset("temp", memory=data)
#                 ds = xr.open_dataset(xr.backends.NetCDF4DataStore(ds0))
#             ds_list.append(ds)

#         return ds_list
#     else:
#         if status == "restore":
#             print(
#                 "[WARNING]: Data for request "
#                 + data_id
#                 + " did not become ready before the max-wait time expired. You can rerun and ask for this request by id"
#             )
#         elif status == "running":
#             print(
#                 "[WARNING]: Data for request "
#                 + data_id
#                 + " is still being constructed when the max-wait time expired. Please check on it later"
#             )
#         elif status == "queued":
#             print(
#                 "[WARNING]: Data for request "
#                 + data_id
#                 + " is still queued. If this does not change soon, please contact an administrator"
#             )
#         else:
#             print("[ERROR]: Data has not become available due to an unknown error")
#         return


# def met_get(**kwargs):
#     """ "based on the get_metget_data.py script but instead of downloading a ntecdf files it returns an xr.dataset
#     in memory"""
#     import getpass
#     import socket
#     import types

#     AVAILABLE_VARIABLES = {"wind_pressure", "rain", "temperature", "humidity", "ice"}
#     AVAILABLE_FORMATS = {"ascii", "owi-ascii", "adcirc-netcdf", "hec-netcdf", "delft3d"}

#     args = types.SimpleNamespace()

#     for name in kwargs.keys():
#         exec('args.{} = kwargs["{}"]'.format(name, name))

#     optional = {
#         "analysis": False,
#         "multiple_forecasts": True,
#         "format": "hec-netcdf",
#         "variable": "wind_pressure",
#         "check_interval": 30,
#         "max_wait": 24,
#         "strict": True,
#         "backfile": True,
#         "epsg": 4326,
#         "dryrun": False,
#         "request": None,
#         "compression": False,
#         "endpoint": None,
#         "apikey": None,
#     }

#     for key in optional.keys():
#         if key not in kwargs:
#             exec('args.{} = optional["{}"]'.format(key, key))

#     if not args.endpoint:
#         if "METGET_ENDPOINT" not in os.environ:
#             raise RuntimeError("No endpoint found.")
#         else:
#             endpoint = os.environ["METGET_ENDPOINT"]
#     else:
#         endpoint = args.endpoint

#     if not args.apikey:
#         if "METGET_API_KEY" not in os.environ:
#             raise RuntimeError("No API key was found.")
#         else:
#             apikey = os.environ["METGET_API_KEY"]
#     else:
#         apikey = args.apikey

#     # ...Check for required arguments
#     if not args.request:
#         if not args.start:
#             print("[ERROR]: Must provide '--start'")
#             exit(1)
#         if not args.end:
#             print("[ERROR]: Must provide '--end'")
#             exit(1)
#         if not args.timestep:
#             print("[ERROR]: Must provide '--timestep'")
#             exit(1)

#         # ...Building the request
#         domains = []
#         idx = 0
#         for d in args.domain:
#             j = parse_domain_data(d, idx)
#             domains.append(j)
#             idx += 1

#         if args.format not in AVAILABLE_FORMATS:
#             print("ERROR: Invalid output format selected")
#             exit(1)

#         if args.format == "delft3d" or args.format == "hec-netcdf":
#             if len(domains) > 1:
#                 print(
#                     "[ERROR]: "
#                     + args.format
#                     + " does not support more than one domain."
#                 )
#                 exit(1)

#         if args.variable not in AVAILABLE_VARIABLES:
#             print("ERROR: Invalid variable selected")
#             exit(1)

#         request_from = getpass.getuser() + "." + socket.gethostname()
#         request_data = {
#             "version": "0.0.1",
#             "creator": request_from,
#             "background_pressure": 1013.0,
#             "backfill": True,
#             "nowcast": args.analysis,
#             "multiple_forecasts": args.multiple_forecasts,
#             "start_date": str(args.start),
#             "end_date": str(args.end),
#             "format": args.format,
#             "data_type": args.variable,
#             "time_step": args.timestep,
#             "domains": domains,
#             "compression": args.compression,
#             "epsg": args.epsg,
#             "filename": args.output,
#         }
#         if args.strict:
#             request_data["strict"] = True
#         if args.dryrun:
#             request_data["dry_run"] = True

#         data_id, status_code = make_metget_request(endpoint, apikey, request_data)
#         if not args.dryrun and status_code == 200:
#             ds_list = download_metget_data(
#                 data_id, endpoint, apikey, args.check_interval, args.max_wait
#             )
#         else:
#             print(status_code)
#             return None

#     else:
#         ds_list = download_metget_data(
#             args.request, endpoint, apikey, args.check_interval, args.max_wait
#         )

#     return ds_list



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""
import os
import datetime
from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore
import xarray as xr
import pandas as pd
import numpy as np

from .dataset import MeteoDataset

class MeteoDatasetGFSForecast0p25(MeteoDataset):

    # Inherit from MeteoDomain
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set some source information
        self.source_name = "gfs_forecast_0p25"
        self.source_type = "forecast"
        self.source_delay = 6
        self.source_cycle_interval = 6
        self.source_time_interval = 3
    
    def download_forecast_cycle(self, **kwargs):
        """Downloads COAMPS-TC forecast cycle for a given storm number and cycle time"""

        if "cycle_time" in kwargs:
            cycle_time = kwargs["cycle_time"]
        else:
            # Throw error if cycle_time is not provided            
            print("Error: cycle_time not provided")
            return

        if "time_range" in kwargs:
            time_range = kwargs["time_range"]
        else:
            # Get all data from this cycle
            time_range = [cycle_time, cycle_time + datetime.timedelta(hours=self.source_forecast_duration)]

        cycle_string = cycle_time.strftime("%Y%m%d_%H%M")
        cycle_name   = cycle_time.strftime("%Y%m%d_%Hz")

        # Make folder for the forecast
        forecast_path = os.path.join(self.path, cycle_name) # Folder to store the forecast netcdf files

        # Make folder for the forecast
        os.makedirs(forecast_path, exist_ok=True) 

        base_url = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/"
        url = base_url + "GFS_Global_0p25deg_" + cycle_string + ".grib2/catalog.xml"
        url = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/catalog.xml?dataset=grib/NCEP/GFS/Global_0p25deg/Best"
        # TODO right now the url is replaced to use the best (where we don't know the cycle used)
        # TODO what about using S3 bucket (e.g. https://noaa-gfs-bdp-pds.s3.amazonaws.com/index.html#gfs.20240627/)
        
        # We assume that the best uses the latest
        latest_xml  = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/latest.xml"
        latest_file = TDSCatalog(latest_xml).catalog_name
        latest_date = "".join(latest_file.split(".")[0].split("_")[-2:])
        latest_cycle_time = datetime.datetime.strptime(latest_date, "%Y%m%d%H%M").replace(tzinfo=datetime.timezone.utc)
        # If we request times later than the latest available cycle the latest has been used
        if time_range[0] > latest_cycle_time:
            cycle_time = latest_cycle_time
        
        gfs   = TDSCatalog(url)
        ncss    = list(gfs.datasets.values())[0].subset()

        param_list = ["wind", "barometric_pressure", "precipitation"]

        ds = xr.Dataset()

        # Loop through requested parameters     
        for param in param_list:
    
            if param == "wind":
                
                # dataset.quantity = param
                
                query = ncss.query()
                query.lonlat_box(north=self.lat_range[1], \
                                south=self.lat_range[0], \
                                east=self.lon_range[1],  \
                                west=self.lon_range[0]).time_range(time_range[0], time_range[1]).vertical_level(10.0)
                query.variables("u-component_of_wind_height_above_ground",
                                "v-component_of_wind_height_above_ground")
                ncss_data = ncss.get_data(query)
                ds0 = xr.open_dataset(NetCDF4DataStore(ncss_data))

                lat = np.array(ds0["latitude"])                 
                lat = np.flip(lat)

                ds["lon"] = np.array(ds0["longitude"]) - 360.0
                ds["lat"] = lat
                ds["time"] = ds0["time"]

                ds["wind_u"] = xr.DataArray(np.flip(np.squeeze(ds0['u-component_of_wind_height_above_ground'].values)), dims=("time", "lat", "lon"))
                ds["wind_v"] = xr.DataArray(np.flip(np.squeeze(ds0['v-component_of_wind_height_above_ground'].values)), dims=("time", "lat", "lon"))

                ds0.close()
            
            else:
                
                # Other scalar variables                
                fac = 1.0

                if param == "barometric_pressure":
                    var_name = "Pressure_reduced_to_MSL_msl"
                elif param == "precipitation":    
                    var_name = "Precipitation_rate_surface"
                    fac = 3600.0
            
                query = ncss.query()            
                query.lonlat_box(north=self.lat_range[1], \
                                south=self.lat_range[0], \
                                east=self.lon_range[1],  \
                                west=self.lon_range[0]).time_range(time_range[0], time_range[1])
                query.variables(var_name)
                ncss_data = ncss.get_data(query)
                ds0 = xr.open_dataset(NetCDF4DataStore(ncss_data))

                # Check if lon, lat and time are already in the dataset
                if "lon" not in ds or "lat" not in ds or "time" not in ds:
                    ds["lon"] = np.array(ds0["longitude"]) - 360.0
                    lat = np.array(ds0["latitude"])                 
                    lat = np.flip(lat)
                    ds["lat"] = lat
                    ds["time"] = ds0["time"]

                v = np.flip(np.squeeze(ds0[var_name].values))

                if param == "precipitation":
                    v = v * fac

                ds[param] = xr.DataArray(v, dims=("time", "lat", "lon"))

                ds0.close()

        write2nc(ds, self.name, os.path.join(self.path, cycle_name))

        ds.close()

        self.ds = ds

def write2nc(ds, meteo_name, meteo_path):

    # Loop though times in ds
    times = ds["time"].values
    for it, t in enumerate(times):        
        time_string = pd.to_datetime(t).strftime("%Y%m%d_%H%M")    
        file_name = meteo_name + "." + time_string + ".nc"
        full_file_name = os.path.join(meteo_path, file_name)        
        ds_time = ds.isel(time=it)
        # Remove time and reftime
        ds_time = ds_time.drop_vars(["time", "reftime"])
        ds_time.to_netcdf(path=full_file_name)
        ds_time.close()



# -*- coding: utf-8 -*-
"""
Created on Tue May 18 12:00:56 2021

@author: ormondt
"""

import glob
import os
import shutil


def move_file(src, dst):
    for full_file_name in glob.glob(src):
        src_name = os.path.basename(full_file_name)
        if os.path.exists(os.path.join(dst, src_name)):
            # Try removing existing file
            try:
                os.remove(os.path.join(dst, src_name))
            except Exception:
                print("Could not remove file " + os.path.join(dst, src_name))
        try:
            shutil.move(full_file_name, dst)
        except Exception:
            print("Could not move file " + full_file_name)


def copy_file(src, dst):
    for full_file_name in glob.glob(src):
        src_name = os.path.basename(full_file_name)
        if os.path.exists(os.path.join(dst, src_name)):
            os.remove(os.path.join(dst, src_name))
        if os.path.isdir(full_file_name):
            dstf = os.path.join(dst, os.path.basename(full_file_name))
            shutil.copytree(full_file_name, dstf)
        else:
            shutil.copy(full_file_name, dst)


def delete_file(src):
    for file_name in glob.glob(src):
        try:
            os.remove(src)
        except Exception:
            print("Could not delete " + src)


def rm(src):
    os.remove(src)


def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)


def list_files(src, full_path=True):
    if full_path:
        file_list = []
        full_list = glob.glob(src)
        for it, item in enumerate(full_list):
            if os.path.isfile(item):
                file_list.append(item)
    else:
        file_list = os.listdir(src)
    #                file_list.append(os.path.basename(item))
    #                print(it)

    return sorted(file_list)


def list_folders(src, basename=False):
    folder_list = []
    full_list = glob.glob(src)
    for item in full_list:
        if os.path.isdir(item):
            if basename:
                folder_list.append(os.path.basename(item))
            else:
                folder_list.append(item)

    return sorted(folder_list)


def delete_folder(src):
    try:
        shutil.rmtree(src, ignore_errors=False, onerror=None)
    except Exception:
        # Folder was probably open in another application
        print("Could not delete folder " + src)


def rmdir(src):
    try:
        if os.path.exists(src):
            shutil.rmtree(src, ignore_errors=False, onerror=None)
    except Exception:
        # Folder was probably open in another application
        print("Could not delete folder " + src)


def exists(src):
    if os.path.exists(src):
        return True
    else:
        return False



# -*- coding: utf-8 -*-
from metget.metget_build import MetGetBuildRest
import tomli
import pandas as pd
import numpy as np
from pyproj import CRS
import datetime
from cht_meteo.cht.meteo import gfs_forecast_0p25
from .coamps_utils import *
class Dataset():

    def __init__(self):

        self.quantity = None
        self.unit     = None
        self.time     = []
        self.x        = None
        self.y        = None
        self.crs      = None
        self.val      = None
        self.source   = []


def download(param_list, lon_range, lat_range, time_range, cycle_time, resolution=0.25, config_path=None):
    """Function to download coamps-tc forecasts using the metget tool api
    Right now resolution is hardcoded to 0.25 degrees but this can change"""
    # TODO gfs should not be downloaded in this script! All this logic should be in cosmos!
    try:
        fill_values = dict(wind=0.0, precipitation=0.0, barometric_pressure=102000.0)  # fill values to use in case of nan
        units = dict(wind='m/s', precipitation='kg.m-2.hour-1', barometric_pressure='Pa')  # units

        if not config_path:
            raise ValueError("The path to a configuration file needs to be provided with the argument 'config_path'")
        else:
            with open(config_path, mode="rb") as fp:
                config = tomli.load(fp)
                
        # Read data from the config file
        apikey = config["apikey"]
        endpoint = config["endpoint"]
        api_version = config["api_version"]
        if "tau" in config:
            tau = config["tau"]
        else:
            tau = 0

        # Connect parameters names with coamps-tc names
        param_names = {'wind': 'wind_pressure', 'barometric_pressure': 'wind_pressure', 'precipitation': 'rain'}
        variables = list(np.unique([param_names[name] for name in param_list])) # ['wind_pressure', 'rain']

        # Get individual times requested
        # TODO how is time-step handled in cosmos?
        requested_times = pd.date_range(start=time_range[0],
                        end=time_range[1],
                        freq='3H').to_pydatetime().tolist()
        requested_times = [ti.replace(tzinfo=datetime.timezone.utc) for ti in requested_times]

        timestep = int((requested_times[1] - requested_times[0]).total_seconds())  # get time-step of requested time assuming that the time-step stays the same
        ntime = len(requested_times)

        # Get the available storms in the Coamps-TC forecast
        try:
            coamps_storms = check_coamps(apikey, endpoint, time_range[0])
        except:
            coamps_storms = {}
        
        # Get the storm names that have a forecast available for the requested cycle
        storms = [name for name in coamps_storms.keys() if cycle_time.strftime("%Y-%m-%d %H:%M:%S") in coamps_storms[name]["cycles_complete"]]

        # If there are no storms available raise Error
        if len(storms) == 0:
            print(f"There are no COAMPS-TC forecasts available for cycle: {cycle_time.strftime('%Y-%m-%d %H:%M:%S')}! Using GFS instead")
            datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
            return datasets

        # Check if there is a priority storm given
        if "priority_storm" in config:
            if config["priority_storm"] == "tc_vitals":
                # Get storm priority from noaa
                priority_storm = tc_vitals_storm()
                if not priority_storm:
                    print("Priority storm could not be loaded from https://ftp.nhc.noaa.gov/atcf/com/tcvitals")
                else:
                    print(f"Priority storm {priority_storm} found from https://ftp.nhc.noaa.gov/atcf/com/tcvitals")
            else:
                priority_storm = config["priority_storm"]
                print(f"Priority storm {priority_storm} provided in config file.")  
        else:
            priority_storm=None
        
        if priority_storm is not None and config["priority_storm"] != "tc_vitals" and priority_storm not in storms:
            print(f"Priority storm {priority_storm} not found in available COAMPS-TC forecasted storms {storms}.") 
            datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
            print("Could not find any coamps-tc data in requested time range! Using GFS instead")
            return datasets
        
        if priority_storm not in storms:
            print(f"Priority storm {priority_storm} not found in available COAMPS-TC forecasted storms {storms}.") 
            priority_storm = None
        # Choose storm according to input
        if not priority_storm and len(storms) == 1: # if only a single storm is available
            storm = storms[0]
        elif not priority_storm and len(storms) > 1: # if multiple storms are available use latest
            ids = [int(i.split("L")[0]) for i in storms]
            ind = np.argmax(ids)
            storm = storms[ind]
        else:
            storm = priority_storm
        
        print(f"Forecast will be downloaded for storm {storm}")  

        # Prepare metget domain
        domain = [f"coamps-{storm}"] + [resolution] + [lon_range[0], lat_range[0], lon_range[1], lat_range[1]]
        t1 = requested_times[0]
        t2 = requested_times[-1]
        
        dss, metadata = {}, {}
        
        for ii, var in enumerate(variables):
            # ...Building the request
            request_data = MetGetBuildRest.generate_request_json(
                start_date=t1.strftime("%Y%m%d %H%M%S"),
                end_date=t2.strftime("%Y%m%d %H%M%S"),
                format='hec-netcdf',
                data_type=var,
                time_step=timestep,
                domains=MetGetBuildRest.parse_command_line_domains([domain], tau),
                epsg=4326,
                filename=f"{storm}_{var}",
                backfill=True,
                # nowcast=False,
                multiple_forecasts=False, # This makes sure that we only use a single forecasting cycle
                # compression=True,
                save_json_request=False,
                # dry_run=True,
                strict=True,
            )
            client = MetGetBuildRest(endpoint, apikey, api_version)
            data_id, status_code = client.make_metget_request(request_data) # Status code should be 200 if everything is ok
            if status_code != 200:
                print(f"metget request return a status code of {status_code}. Data could not be downloaded. Using GFS instead") 
                datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
                return datasets
            urls, meta = client.download_metget_data(
                data_id,
                30, # sleep_time in seconds
                3, # max_wait in hours
                output_directory=None,
                return_only_url=True,
            )
            # TODO Check why this should be a list?
            ds_list = []
            for url in urls:
                ds = get_da_from_url(url)
                ds_list.append(ds)
            if ds_list is not None:
                dss[var] = ds_list[0]
                metadata[var] = meta
            else:
                dss[var] = None

        # if no data could be downloaded (CHECK AGAIN WHAT SHOULD HAPPEN IN THIS CASE)
        if all([dss[k] is None for k in list(dss.keys())]):
            datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
            print("Could not find any coamps-tc data in requested time range! Using GFS instead")
            return datasets
        else: # Else if there are available data get lon lat info
            for key in dss.keys():
                if dss[key] is not None:
                    data0 = dss[key]
                    lon = np.array(data0['lon'])
                    lat = np.array(data0['lat'])
                    if lat[1] - lat[0] > 0:  # lat should be in descending order
                        lat = lat[::-1]
                        reverse = True
                    else:
                        reverse = False
                    lon = (lon + 360) % 360  # added this to be consistent with gfs
                    nrows = len(lat)
                    ncols = len(lon)

        # Prepare datasets
        datasets = []
        for param in param_list:
            dataset = Dataset()
            dataset.crs = CRS.from_epsg(4326)
            dataset.quantity = param
            dataset.x = lon
            dataset.y = lat
            dataset.time = pd.to_datetime([tt.replace(tzinfo=None) for tt in requested_times]).to_pydatetime()
            if dataset.quantity == "wind":
                dataset.u    = np.empty((ntime, nrows, ncols))
                dataset.u[:] = np.NaN
                dataset.v    = np.empty((ntime, nrows, ncols))
                dataset.v[:] = np.NaN
            else:
                dataset.val    = np.empty((ntime, nrows, ncols))
                dataset.val[:] = np.NaN
            datasets.append(dataset)


        for it, time_i in enumerate(requested_times):
            # Loop through requested parameters
            for ind, param in enumerate(param_list):
                dataset = datasets[ind]
                okay = False
                makezeros = False
                #  First check if there are coamps-tc data available for this parameter
                if dss[param_names[param]] is not None:
                    okay = True
                    model_t = pd.to_datetime(dss[param_names[param]].time.values).to_pydatetime()
                    model_t_ind = np.where(model_t == time_i.replace(tzinfo=None))[0][0]
                else:
                    makezeros = True

                if okay:
                    # Get metadata of input used
                    dataset.config = metadata[param_names[param]]["input"]
                    naming_format = metadata[param_names[param]]["input_files"][f"coamps-tc-{storm}"][0]
                    # Check naming of files to get information on cycle used
                    if naming_format.split("_")[0] == "coamps-tc":
                        cycles_used = [name.split("_")[2] for name in metadata[param_names[param]]["input_files"][f"coamps-tc-{storm}"]]
                        taus = [name.split("_")[3].split(".")[0].split("tau")[1] for name in metadata[param_names[param]]["input_files"][f"coamps-tc-{storm}"]]
                        times = [datetime.datetime.strptime(cycle, "%Y%m%d%H") + datetime.timedelta(hours=int(tau)) for (cycle, tau) in zip(cycles_used, taus)]
                        times = [t.strftime("%Y%m%d%H%M") for t in times]
                        inputs = [name for i, name in enumerate(metadata[param_names[param]]["input_files"][f"coamps-tc-{storm}"]) if time_i.strftime("%Y%m%d%H%M") in times[i]]
                    else:
                        datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
                        print("Could not read cycle used in metget metadata! Using GFS instead")
                        return datasets 

                    # get cycle info
                    cycle_used = inputs[0].split("_")[2]
                    cycle_hour = inputs[0].split("_")[3].split(".")[0].split("tau")[1]
                    dataset.source.append(f'coamps_tc_{storm}_{cycle_used}z_{cycle_hour}')
                    
                    if param == "wind":
                        data = dss[param_names[param]]
                        u = data['wind_u']
                        v = data['wind_v']
                        dataset.unit = u.units
                        
                        if np.any(np.isnan(u)) or np.any(np.isnan(v)):  # check if there are nans and fill them up
                            u = u.fillna(fill_values[param])
                            v = v.fillna(fill_values[param])
                        
                        u = u.metpy.unit_array.squeeze()
                        v = v.metpy.unit_array.squeeze()

                        if reverse:
                            dataset.u[it,:,:] = np.array(u[model_t_ind, ::-1, :])
                            dataset.v[it,:,:] = np.array(v[model_t_ind, ::-1, :])
                        else:
                            dataset.u[it,:,:] = np.array(u[model_t_ind, :, :])
                            dataset.v[it,:,:] = np.array(v[model_t_ind, :, :])
                    else:
                        if param == "barometric_pressure":
                            var_name = "mslp"
                        elif param == "precipitation":
                            var_name = 'precipitation'
                        data = dss[param_names[param]]
                        val          = data[var_name]

                        # Added this check to ensure that pressure is in Pa
                        if param == "barometric_pressure":
                            if val.units == 'mb':
                                val = val * 100
                                val.attrs['units'] = 'Pa'
                        dataset.unit = val.units

                        if param == "precipitation":  # Added this check to ensure that nan in precipitation are correctly interpreted
                            val = val.where(val>=0, np.nan)

                        if np.any(np.isnan(val)):  # check if there are nans and fill them up
                            val = val.fillna(fill_values[param])
                            
                        val = np.array(val.metpy.unit_array.squeeze())

                        if reverse:
                            dataset.val[it, :, :] = np.array(val[model_t_ind, ::-1, :])
                        else:
                            dataset.val[it, :, :] = np.array(val[model_t_ind, :, :])

                elif makezeros: # add zeros
                    try:
                        #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
                        datasets_gfs = gfs_forecast_0p25.download(param_list, lon_range, lat_range,
                                                                [time_i, time_i], cycle_time)
                        if param == "wind":
                            dataset.u[it, :, :] = datasets_gfs[ind].u
                            dataset.v[it, :, :] = datasets_gfs[ind].v
                        else:
                            dataset.val[it, :, :]  = datasets_gfs[ind].val
                        dataset.unit = datasets_gfs[ind].unit
                        dataset.src.append('gfs')
                        print(param + " was not found on server for {} --> Using GFS forecast data instead".format(time_i))
                    except:
                        if param == "wind":
                            dataset.u[:] = fill_values[param]
                            dataset.v[:] = fill_values[param]
                            dataset.unit = units[param]
                            dataset.src.append('default_value')
                            print(param + " was not found on server for {} --> using {} {} instead !!!".format(time_i, fill_values[param], units[param]))

                        else:
                            dataset.val[:] = fill_values[param]
                            dataset.unit = units[param]
                            dataset.src.append('default_value')
                            print(param + " was not found on server ... --> using {} {} instead !!!".format(fill_values[param], units[param]))
    except Exception as e:
        print(e)
        datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
        print("Something went wrong when trying to download coamps-tc data! Using GFS instead")
        return datasets 


    return datasets







# -*- coding: utf-8 -*-
"""
Created on Fri Sep  9 12:01:21 2022

@author: wcp_w2111150
"""

# NOTE YOU NEED GEOPANDAS

import datetime
import matplotlib.pyplot as plt
from cht_meteo.meteo import MeteoSource, MeteoGrid
from cht_meteo.meteo import filter_cyclones_TCvitals

# Define a track
source = MeteoSource("coamps", "coamps_tc_forecast", "forecast",
                     delay=6)

parameters = ["wind", "barometric_pressure"]
x_range = [-100.0, -80.0]
y_range = [10.0, 30.0]
ds = MeteoGrid(name="coamps", source=source, parameters=parameters,
               path="d:\data\COAMPS\hurricane_maria_coamps_new_t4",
               x_range=x_range,
               y_range=y_range)

t0 = datetime.datetime(2017,9,17,0,0,0)
t1 = datetime.datetime(2017,9,22,0,0,0)
ds.collect(time_range=[t0, t1])

# New determine best-track from these files
tracks = ds.find_cyclone_tracks(pcyc=99999, dt=6)

# Filter cyclone based on TCvitals
tracks = filter_cyclones_TCvitals(tracks)

# Write them out as geojson at location
for k, track in enumerate(tracks):
    
    # Check the size - only if there are 3 timestamps
    if track.track.geometry.size > 3:

        # Save as cyc
        output = r'd:\data\COAMPS\maria_tracks' + str(k) + '.cyc'
        track.write_track(output, 'ddb_cyc')

        # Save as shapefile
        output = r'd:\data\COAMPS\maria_tracks' + str(k) + '.shp'
        track.track.to_file(output)


# Use the first track to make ensembles
from cht_cyclones.tropical_cyclone import TropicalCyclone, TropicalCycloneEnsemble
from datetime import datetime, timedelta
import fiona
tc = tracks[0]
tc.account_for_forward_speed()
tc.estimate_missing_values()
tc.include_rainfall = True

tc2         = TropicalCycloneEnsemble(name="test", TropicalCyclone=tc)
tc2.tstart  = datetime(2017,9,17,0,0,0)
tc2.tend    = datetime(2017,9,22,0,0,0)
tc2.compute_ensemble(100)

fname = r'd:\data\COAMPS\ensembles'
tc2.to_shapefile(fname)
#tc2.to_spiderweb(fname)
#tc2.make_figures(fname)

# Done
print('done!')


import os
import datetime
import xarray as xr
import numpy as np
import pandas as pd
from pyproj import CRS
from pyproj import Transformer
from scipy.interpolate import RegularGridInterpolator
import time

from .dataset_to_delft3d import write_to_delft3d_ascii
from .dataset_to_json_wind import write_wind_to_json

import cht_utils.fileops as fo

class MeteoDataset():
    def __init__(self, **kwargs):

        # Default values set to None 
        self.name = None               # Name of the dataset  
        self.path = None               # Path to the dataset
        self.long_name = None          # Long name of the dataset (not really needed)
        self.source = None             # Source of the dataset
        self.parameters = None         # Parameters in the dataset (a list with strings). Not currently used.
        self.lon_range = None          # Longitude range of the dataset
        self.lat_range = None          # Latitude range of the dataset
        self.var_names = ["wind_u", "wind_v", "barometric_pressure", "precipitation"] # Variable names in the dataset
        self.crs = CRS(4326)           # Coordinate reference system of the dataset
        self.tau = 0                   # Time interval in hours between cycle and data
        self.last_analysis_time = None # Time of last analysis in the dataset

        # Loop through kwargs to set attributes
        reserved_keys = ["x", "y", "lon", "lat", "time"]    
        for key, value in kwargs.items():            
            if key not in reserved_keys:
                setattr(self, key, value)  

        # Set some source information
        self.source_name = ""               # Name of the source
        self.source_type = "forecast"       # Can be "forecast" or "analysis" 
        self.source_delay = 0               # Delay in hours before data is available
        self.source_cycle_interval = 6      # Interval in hours between cycles
        self.source_time_interval = 3       # Output interval in hours
        self.source_forecast_duration = 240 # The length of the forecast in hours that will be downloaded

        # Empty list for subsets (only for COAMPS-TC for now)
        self.subset = []

        # Create empty xarray dataset
        self.ds = xr.Dataset()

        time = None
        x = None
        y = None

        if "time" in kwargs:
            time = kwargs["time"]
        if "x" in kwargs:
            x = kwargs["x"]    
        if "lon" in kwargs:
            x = kwargs["lon"]    
        if "y" in kwargs:
            y = kwargs["y"]
        if "lat" in kwargs:
            y = kwargs["lat"]

        if time is not None and x is not None and y is not None: 
            self.init_ds(time, x, y)

    def init_ds(self, time, x, y):
        """Fill the dataset with nodata values."""

        x = np.array(x)
        y = np.array(y)
        time = np.array(time)    

        # Create empty dataset
        self.ds["time"] = xr.DataArray(time, dims=("time"))
        if self.crs.is_geographic:
            self.ds["lon"] = xr.DataArray(x, dims=("lon"))
            self.ds["lat"] = xr.DataArray(y, dims=("lat"))
        else:
            self.ds["x"] = xr.DataArray(x, dims=("x"))
            self.ds["y"] = xr.DataArray(y, dims=("y"))
        empty_data = np.empty((len(time), len(y), len(x))) + np.nan  
        for var_name in self.var_names:
            if self.crs.is_geographic:
                self.ds[var_name] = xr.DataArray(empty_data, dims=("time", "lat", "lon"))    
            else:
                self.ds[var_name] = xr.DataArray(empty_data, dims=("time", "y", "x"))

    def download(self, time_range, **kwargs):

        # Make path
        os.makedirs(self.path, exist_ok=True)

        if self.source_type == "forecast":
            # Download forecast from cycles
            self.download_forecast(time_range, **kwargs)
        else:
            # Download analysis
            self.download_analysis(time_range, **kwargs)

    def download_forecast(self, time_range, **kwargs):            # Need to check on previous cycles
        """"Download all (!) data from a forecast dataset. This is done by downloading all cycles within the time range."""

        # Check if this dataset has a download_forecast_cycle method
        if not hasattr(self, "download_forecast_cycle"):
            print(f"Error: download_forecast_cycle method not implemented for dataset {self.name}")
            return

        # Round first time in range down to hour        
        h0 = time_range[0].hour
        # Round down to cycle interval
        h0 = h0 - np.mod(h0, self.source_cycle_interval)
        t0 = time_range[0].replace(microsecond=0, second=0, minute=0, hour=h0,
                                    tzinfo=datetime.timezone.utc)

        # Round last time in range down to hour        
        h1 = time_range[1].hour
        # Round down to cycle interval
        h1 = h1 - np.mod(h1, self.source_cycle_interval)
        t1 = time_range[1].replace(microsecond=0, second=0, minute=0, hour=h1,
                                    tzinfo=datetime.timezone.utc)
        
        print("Now : " + datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%d_%H%M%S"))

        # Determine latest available cycle  
        t_latest = datetime.datetime.now(datetime.timezone.utc) - \
            datetime.timedelta(hours=self.source_delay)
        h0 = t_latest.hour
        h0 = h0 - np.mod(h0, self.source_cycle_interval)
        t_latest = t_latest.replace(microsecond=0, second=0, minute=0, hour=h0)
        print("t_latest : " + t_latest.strftime("%Y%m%d_%H%M%S"))

        # Make sure t0 and t1 are not later than t_latest        
        t0 = min(t0, t_latest)
        t1 = min(t1, t_latest)

        print(f"Downloading from {self.name} - cycles : " + t0.strftime("%Y%m%d_%Hz") + " to " + t1.strftime("%Y%m%d_%Hz"))

        # Make list with cycle times
        cycle_times = pd.date_range(start=t0,
                                    end=t1,
                                    freq=str(self.source_cycle_interval) + 'h').to_pydatetime().tolist()
        
        # Loop through all cycle times
        for it, t in enumerate(cycle_times):
            print(f"Downloading {it + 1} of {len(cycle_times)} - cycle : " + t.strftime("%Y%m%d_%Hz"))
            try:
                self.download_forecast_cycle(cycle_time=t,
                                             **kwargs)
            except Exception as e:
                print(f"Error downloading data from dataset {self.name} - cycle : {t.strftime('%Y%m%d_%Hz')}")
                print(e)

    def download_analysis(self, time_range, **kwargs):

        # Check if this dataset has a download_forecast_cycle method
        if not hasattr(self, "download_analysis_times"):
            print(f"Error: download_analysis_times method not implemented for dataset {self.name}")
            return

        # Make list of requested times
        freqstr = str(self.source_time_interval) + 'h'
        requested_times = pd.date_range(start=time_range[0],
                            end=time_range[1],
                            freq=freqstr).to_pydatetime().tolist()
        
        # Loop through all requested times and see which data is already available
        rtimes = []
        # Check which files do not yet exist
        for t in requested_times:
            file_name = os.path.join( self.path, f"{ self.name }.{ t.strftime('%Y%m%d_%H%M')}.nc" )
            if not os.path.exists(file_name):
                rtimes.append(t)                  
        
        if rtimes:
            try:
                self.download_analysis_times(rtimes)
            except Exception as e:
                print(f"Error downloading data from dataset {self.name}")
                print(e)
        else:
            print("Requested meteo data already available")

    def collect(self, time_range, **kwargs):
        """Merge data from separate netcdf files. The actual data is stored in the xarray dataset self.ds. This is a list, so we can optionally store multiple subsets of the data."""

        if "tau" in kwargs:
            tau = kwargs["tau"]
        else:
            tau = self.tau

        last_cycle_time = None
        if "last_cycle" in kwargs:
            if kwargs["last_cycle"] is not None:
                # last_cycle_time = datetime.datetime.strptime(kwargs["last_cycle"], "%Y%m%d_%H")
                last_cycle_time = kwargs["last_cycle"]

        # Subsets are only used when there are subsets with different resolutions (e.g. as in COAMPS-TC) 
        if len(self.subset) > 0:
            # There are subsets in this dataset. For now we automatically collect data of each subset.
            subsets = True
            subsets_to_get = []
            for subset in self.subset:
                subsets_to_get.append(subset.name)

        else:
            subsets = False
            subsets_to_get = ["dummy"]        

        # Loop through all subsets (i.e. nesting levels)
        for isub, subset_name in enumerate(subsets_to_get):

            # Make empty lists for time and files  
            time_list = []
            file_list = []

            if subsets:
                subsetstr = "." + self.subset[isub].name + "."
                moving = self.subset[isub].moving
            else:
                subsetstr = ""
                moving = False    
         
            if self.source_type == "forecast":

                # Make list of all cycles
                all_cycle_paths = fo.list_folders(os.path.join(self.path, "*"))

                icycle = -1

                # Loop through all cycle paths
                for cycle_path in all_cycle_paths:

                    # Get time from cycle path
                    t_cycle = datetime.datetime.strptime(cycle_path[-12:-1], "%Y%m%d_%H")

                    # If last_cycle has been provided, do not get data from later cycles
                    if last_cycle_time is not None:
                        if t_cycle > last_cycle_time:
                            # We should stop now
                            break

                    # Check if time of this cycle falls within requested range
                    if t_cycle < time_range[0] or t_cycle > time_range[1]:
                        # We can skip this cycle
                        continue

                    # Find all times available in this cycle as it may contain our data
                    files_in_cycle = fo.list_files(os.path.join(cycle_path, "*" + subsetstr + "*.nc"))

                    icycle += 1

                    # Loop through all files in this cycle
                    for ifile, file in enumerate(files_in_cycle):

                        t_file = datetime.datetime.strptime(file[-16:-3],
                                                            "%Y%m%d_%H%M")
                        
                        if ifile == 0:
                            self.last_analysis_time = t_file

                        if tau > 0 and icycle > 0:                            
                            # Compute time interval between cycle and file
                            th = int((t_file - t_cycle).total_seconds() / 3600)
                            if th < tau:
                                # We can skip this file
                                continue

                        if t_file < time_range[0] or t_file > time_range[1]:
                            # We can skip this file
                            continue

                        # We want the data in this file !

                        # Check if time is already available in time_list. If so, insert it in the correct place. If not, append it.
                        if t_file in time_list:
                            # Find index
                            ind = time_list.index(t_file)
                            time_list[ind] = t_file
                            file_list[ind] = file
                        else:
                            time_list.append(t_file)
                            file_list.append(file)    

                # fill_missing_at_start = True
                # if fill_missing_at_start:
                #     # Use earliest forecast file available for spinup (force earlier timesteps with this file)
                #     id1 = next(i for i,v in enumerate(requested_files) if v is not None)
                #     for ind in range(id1):
                #         requested_files[ind] = requested_files[id1]

                # # Get rid of None values
                # times_to_remove = []
                # for ind, file in enumerate(requested_files):
                #     if not file:
                #         times_to_remove.append(requested_times[ind])
                # if times_to_remove:  
                #     for tr in times_to_remove:
                #         requested_times.remove(tr)
                # requested_files = [value for value in requested_files if value != None]

                # # Turn time array into nump array
                # requested_times = np.array(requested_times)

            else:

                # A lot easier
                files_in_cycle = fo.list_files(os.path.join(self.path, "*" + subsetstr + "*.nc"))
                for file in files_in_cycle:
                    t_file = datetime.datetime.strptime(file[-16:-3],
                                                        "%Y%m%d_%H%M")
                    if t_file>=time_range[0] and t_file<=time_range[1]:
                        file_list.append(os.path.join(self.path, file))
                        time_list.append(t_file)
                        
            # Now we loop through the files, read them and store them in large array
            time = np.array(time_list)
            
            if not time_list:
                print("No meteo data files found within requested time range")
                return

            ntime = len(time_list)

            # Read in first file to get lons and lats
            ds0 = xr.open_dataset(file_list[0])
            lon = ds0["lon"].values[:]
            if lon[0]>180.0:
                lon = lon - 360.0
            lat = ds0["lat"].values[:]

            # Create new dataset
            ds = xr.Dataset()
            # Add time dimension
            ds["time"] = xr.DataArray(time, dims=("time"))
            # Add the lon and lat
            if moving:
                lons = np.empty((ntime, len(lon)))
                lats = np.empty((ntime, len(lat)))
                ds["lon"] = xr.DataArray(lons, dims=("time", "lon"))
                ds["lat"] = xr.DataArray(lats, dims=("time", "lat"))
            else:
                ds["lon"] = xr.DataArray(lon, dims=("lon"))
                ds["lat"] = xr.DataArray(lat, dims=("lat"))

            # First we create empty data arrays
            for var in self.var_names:
                if var == "wind_u" or var == "wind_v" or var == "precipitation":
                    vdefault = 0.0
                elif var == "barometric_pressure":
                    vdefault = 101300.0
                v = np.zeros((ntime, len(lat), len(lon))) + vdefault    
                ds[var] = xr.DataArray(v, dims=("time", "lat", "lon"))

            # Now loop through times
            for it, file in enumerate(file_list):
                # Read in file
                dsin = xr.open_dataset(file)
                if moving:
                    lon = dsin["lon"].values
                    lat = dsin["lat"].values
                    ds["lon"][it,:] = lon
                    ds["lat"][it,:] = lat
                for var in self.var_names:
                    if var in dsin:
                        # Get the data
                        ds[var][it,:,:] = dsin[var].values

            if subsets:
                # Store the data in the subset
                self.subset[isub].ds = ds
            else:
                self.ds = ds

        return ds

    def cut_out(self,
                name=None,
                lon_range=None,
                lat_range=None,
                time_range=None,
                x_range=None,
                y_range=None,
                lon=None,
                lat=None,
                x=None,
                y=None,
                dx=None,
                dy=None,
                copy_time=False,
                crs=None):
        """Returns a new meteo dataset with a cut-out of this dataset. If the original dataset has subsets, these will first be interpolated onto to largest (coarsest) subset."""

        # Start with making a copy of self

        if crs is None:
            crs = self.crs    

        # Check the options (call everything x and y for now)
        if lon_range is not None:
            x_range = lon_range
        if lat_range is not None:
            y_range = lat_range
        if lon is not None:
            x = lon
        if lat is not None:
            y = lat

        # There are now 3 options for the cut-out:
        # 1) none are given (no cut-out, but still interpolation), so set x_range and y_range and continue to option 2
        # 2) x_range and y_range are given, but not dx and dy (just cut-out, no interpolation) - this will only work if the CRS is the same for both datasets
        # 3) x_range and y_range are given, and dx and dy are given (cut-out and interpolate)
        # 4) x and y are given, but not x_range and y_range (cut-out and interpolate)
        
        # Option 1
        if x_range is None and y_range is None and dx is None and dy is None and x is None and y is None:
            # No cut-out, just interpolation
            x_range = [-1.0e9, 1.0e9]
            y_range = [-1.0e9, 1.0e9]

        # Option 2
        if x_range is not None and y_range is not None and dx is None and dy is None:
            # Cut-out only (but possible interpolation)
            if crs != self.crs:
                print("Error: cut-out with different CRS requires arrays x and y, or dx and dy in combination with x_range and y_range")
                return
            # Get coordinates (in case of subsets, we take the largest (coarsest) subset)
            x, y = self.get_coordinates(0)
            # Let's see if x and y are regularly spaced
            if np.min(np.diff(x)) == np.max(np.diff(x)) and np.min(np.diff(y)) == np.max(np.diff(y)):
                # Regular spacing, all good
                pass
            else:
                # Irregular spacing
                # Create new regular x and y arrays with similar spacing
                dx = np.mean(np.diff(x))
                dy = np.mean(np.diff(y))
                x = np.arange(x[0], x[-1], dx)
                y = np.arange(y[0], y[-1], dy)

            # Limit x, y
            ix0 = np.where(x<=x_range[0])[0]
            if len(ix0) > 0:
                ix0 = ix0[-1]
            else:
                ix0 = 0

            ix1 = np.where(x>=x_range[1])[0]
            if len(ix1) > 0:
                ix1 = ix1[0]
            else:
                ix1 = len(x)

            iy0 = np.where(y<=y_range[0])[0]
            if len(iy0) > 0:
                iy0 = iy0[-1]
            else:
                iy0 = 0

            iy1 = np.where(y>=y_range[1])[0]
            if len(iy1) > 0:
                iy1 = iy1[0]
            else:
                iy1 = len(y)

            if ix1<=ix0 or iy1<=iy0:
                print("Error: cut-out is empty with given x_range and y_range")
                return    

            x = x[ix0:ix1]
            y = y[iy0:iy1]

        # Option 3
        elif x_range is not None and y_range is not None and dx is not None and dy is not None:
            # Create new x and y arrays
            x = np.arange(x_range[0], x_range[1], dx)
            y = np.arange(y_range[0], y_range[1], dy)
        
        # Option 4
        elif x is not None and y is not None:
            # No need to do anything
            pass

        if len(self.subset) > 0:
            t = self.subset[0].ds["time"].values
        else:
            t = self.ds["time"].values

        if time_range is not None:
            # convert values in time_range to np.datetime64
            time_range = [np.datetime64(t) for t in time_range]
            t = t[(t>=time_range[0]) & (t<=time_range[1])]

        # Create new dataset
        dataset = MeteoDataset(name=name, x=x, y=y, time=t, crs=crs)
        dataset.interpolate_dataset(self)

        return dataset

    def interpolate_dataset(self, dataset, copy_time=False):
        """Interpolate data from another dataset to this dataset."""
        # Dimensions have already been set (also time?)
        if copy_time:
            self.ds["time"] = dataset.ds["time"]
        # Get horizontal coordinates
        if "x" in self.ds:
            x = self.ds["x"].values
            y = self.ds["y"].values
        else: 
            x = self.ds["lon"].values
            y = self.ds["lat"].values
        xg, yg = np.meshgrid(x, y)
        # Loop through variables
        for var_name in self.var_names:
            da = self.ds[var_name].copy()
            for it, t in enumerate(self.ds["time"].values):
                # Get data
                v = dataset.interpolate_variable(var_name, t, xg, yg, crs=self.crs)
                # Set points in v equal to points in original data vori where vori already has a value
                vori = da.loc[dict(time=t)].values[:]
                not_nan = np.where(~np.isnan(vori))
                v[not_nan] = vori[not_nan]
                da.loc[dict(time=t)] = v

            self.ds[var_name] = da

    def merge_datasets(self, datasets, **kwargs):
        """Merge datasets. This is useful when we have multiple datasets with different resolutions."""
        # For now we just return the first dataset
        for dataset in datasets:
            self.interpolate_dataset(dataset)
    
    def interpolate_variable(self, var_name: str, t: datetime.datetime, x: np.array, y: np.array, crs=None, fill_missing_data=False):
        """Returns numpy array with interpolated values of quantity at requested_time and lon, lat. If quantity is a vector, the function returns the x-component of the vector. If the quantity is not found, the function returns an array with zeros."""

        # Check shape of x and y. They can be either:
        # 1) 1D arrays with regular spacing (turn into grid)
        # 2) 1D arrays with irregular spacing (treat them as point cloud) 
        # 3) 2D arrays with regular or irregular spacing
        if len(np.shape(x)) == 1 and len(np.shape(y)) == 1:
            # 1D arrays
            # Check if x and y are regularly spaced
            if np.min(np.diff(x)) == np.max(np.diff(x)) and np.min(np.diff(y)) == np.max(np.diff(y)):
                # 1D arrays with regular spacing
                xg, yg = np.meshgrid(x, y)
            else:
                # 1D arrays with irregular spacing
                xg = x
                yg = y
        else:
            # 2D arrays
            xg = x
            yg = y    

        if crs is not None:
            # Transformer for x, y to lon, lat
            transformer = Transformer.from_crs(crs,
                                               self.crs,
                                               always_xy=True)
            xg, yg = transformer.transform(xg, yg)

        # Check if there are subsets 
        if len(self.subset) > 0:
            # There are subsets in this dataset. For now we automatically collect data of each subset.
            subsets = True
            nsub    = len(self.subset)
        else:
            subsets = False
            nsub    = 1

        # Make array of nans
        v = np.empty(np.shape(xg))
        v[:] = np.nan

        # Loop in reverse order to get the highest resolution data first
        for isub in range(nsub - 1, -1, -1):
            
            if subsets:
                ds = self.subset[isub].ds
            else:
                ds = self.ds

            # Check if t is a numpy.datetime64
            if not isinstance(t, np.datetime64):
                t = np.datetime64(t)

            # Get data
            if t in ds["time"].values[:]:
                # Get data at time t
                da = ds[var_name].sel(time=t)
            else:
                # Interpolate data at time t                
                da = ds[var_name].interp(time=t)

            # Get horizontal coordinates
            if self.crs.is_geographic:
                x = da["lon"].values
                y = da["lat"].values
            else:
                x = da["x"].values
                y = da["y"].values

            # Make interpolator
            interp = RegularGridInterpolator((y, x), da.values[:])

            # Find points outside of grid
            iout = np.where((xg<np.min(x)) | (xg>np.max(x)) | (yg<np.min(y)) | (yg>np.max(y)))
            x1 = np.maximum(xg, np.min(x))
            x1 = np.minimum(x1, np.max(x))
            y1 = np.maximum(yg, np.min(y))
            y1 = np.minimum(y1, np.max(y))

            # Find points outside of grid
            v1 = interp((y1, x1))

            # Set values outside of grid to no_data_value
            if fill_missing_data:
                if var_name == "barometric_pressure":
                    no_data_value = 101300.0
                else:
                    no_data_value = 0.0
                v1[iout] = no_data_value
            else:
                v1[iout] = np.nan  

            # Where v is nan, replace with v1
            v[np.isnan(v)] = v1[np.isnan(v)]

        return v

    def to_netcdf(self, **kwargs):
        """Write to netcdf files. This is not implemented yet."""
        if "filename" in kwargs:
            # Write to single file
            self.ds.to_netcdf(path=kwargs["filename"])        
        else:    
            # Write to database
            os.makedirs(self.path, exist_ok=True)
            # Loop through times in ds
            time = self.ds["time"].values
            for it, t in enumerate(time):
                time_string = pd.to_datetime(t).strftime("%Y%m%d_%H%M")
                file_name = self.name + "." + time_string + ".nc"
                full_file_name = os.path.join(self.path, file_name)
                ds_time = self.ds.isel(time=it)
                # Remove time and reftime
                ds_time = ds_time.drop_vars(["time", "reftime"])
                ds_time.to_netcdf(path=full_file_name)
                ds_time.close()

    def to_delft3d(self, file_name, version="1.03", path=None, header_comments=False, refdate=None, parameters=None, time_range=None, format="ascii"):

        if len(self.subset) > 0:
            # There are subsets in this dataset. For now we automatically collect data of each subset.
            print("Warning! Cannot write meteo dataset with subsets to delft3d format ! Please consider merging the subsets first.")
            return

        if format == "ascii":
            write_to_delft3d_ascii(self, file_name, version, path, header_comments, refdate, parameters, time_range)
        # else:
        #     write_to_delft3d_netcdf(self, file_name, version, path, header_comments, refdate, parameters, time_range)

    def wind_to_json(self, file_name, time_range=None, js=True, iref=1):
        write_wind_to_json(self, file_name, time_range=time_range, iref=iref, js=js)

    def get_coordinates(self, *args):
        """Returns the horizontal coordinates of the dataset."""
        if len(self.subset) > 0:
            if len(args) > 0:
                isub = args[0]
            else:
                isub = 0
            ds = self.subset[isub].ds
        else:
            ds = self.ds
        if self.crs.is_geographic:
            x = ds["lon"].values
            y = ds["lat"].values
        else:
            x = ds["x"].values
            y = ds["y"].values
        return x, y 

    def get_times(self, *args):
        """Returns the times of the dataset. In np64."""
        if len(self.subset) > 0:
            if len(args) > 0:
                isub = args[0]
            else:
                isub = 0
            ds = self.subset[isub].ds
        else:
            ds = self.ds
        t = ds["time"].values
        return t




# -*- coding: utf-8 -*-
"""
@author:    Panos Athanasiou
Date:       04/07/2022
"""

import datetime
import json
import os
import time

import numpy as np
import pandas as pd
import requests
from pyproj import CRS

from cht_meteo import gfs_forecast_0p25

class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = []
        self.x = None
        self.y = None
        self.crs = None
        self.val = None
        self.src = []


def download(param_list, lon_range, lat_range, time_range, cycle_time, resolution=0.25):
    """Function to download coamps-tc forecasts using the scrubber tool functions
    Right now resolution is hardcoded to 0.25 degrees but this can change"""

    fill_values = dict(
        wind=0.0, precipitation=0.0, barometric_pressure=102000.0
    )  # fill values to use in case of nan
    units = dict(
        wind="m/s", precipitation="kg.m-2.hour-1", barometric_pressure="Pa"
    )  # units

    apikey = "GrPBTsclg71qgDvw4ojGf4mTwZmkbqNu6jcq7T9V"  # Deltares apikey hardcoded for the time being
    endpoint = "https://api.metget.zachcobell.com"

    # Connect parameters names with coamps-tc names
    param_names = {
        "wind": "wind_pressure",
        "barometric_pressure": "wind_pressure",
        "precipitation": "rain",
    }
    variables = list(
        np.unique([param_names[name] for name in param_list])
    )  # ['wind_pressure', 'rain']

    requested_times = (
        pd.date_range(start=time_range[0], end=time_range[1], freq="3H")
        .to_pydatetime()
        .tolist()
    )

    timestep = int(
        (requested_times[1] - requested_times[0]).total_seconds()
    )  # get time-step of requested time assuming that the time-step stays the same
    ntime = len(requested_times)

    # Get storms from endpoint metadata
    storms = check_coamps(apikey, endpoint)

    # Check if forecast exists for the selected dates and make a dataframe
    fr = pd.DataFrame(index=requested_times)
    fr["storm_id"] = None

    for i, ti in enumerate(requested_times):
        fr.loc[ti, "storm_id"] = []
        for j, storm_id in enumerate(storms):
            # The MetGet tool seems to have changed ...
            storm = storms[storm_id]
            ####### CHECK WHICH times to use for range ####################
            t1 = datetime.datetime.strptime(
                storm["min_forecast_date"], "%Y-%m-%d %H:%M:%S"
            )
            t2 = datetime.datetime.strptime(
                storm["max_forecast_date"], "%Y-%m-%d %H:%M:%S"
            )
            # t1 = datetime.datetime.strptime(storm['latest_complete_forecast_start'], "%Y-%m-%d %H:%M:%S")
            # t2 = datetime.datetime.strptime(storm['latest_complete_forecast_end'], "%Y-%m-%d %H:%M:%S")
            belongs = (
                t1.replace(tzinfo=datetime.timezone.utc)
                <= ti.replace(tzinfo=datetime.timezone.utc)
                <= t2.replace(tzinfo=datetime.timezone.utc)
            )
            if belongs:
                fr.loc[ti, "storm_id"].append(storm_id)

    # Check if there is a txt file with the priority storm
    file_storm = r"p:\11206085-onr-fhics\03_cosmos\Priority_Storm.txt"  # !!! Hard-coded
    if os.path.exists(file_storm):
        with open(file_storm, "r") as f:
            lines = f.readlines()
        f.close()
        pr_st_txt = lines[0]
    else:
        pr_st_txt = ""
    # Get storm priority from noaa
    pr_st_noaa = get_priority_storm()
    method = "first"
    for i, ti in enumerate(requested_times):
        if not fr.loc[
            ti, "storm_id"
        ]:  # if there is not coamps storm make None so gfs is used
            fr.loc[ti, "storm_id"] = None
        else:
            if (
                pr_st_txt in fr.loc[ti, "storm_id"]
            ):  # if the priority storm is available that timestep use it
                fr.loc[ti, "storm_id"] = pr_st_txt
                continue
            if (
                pr_st_noaa in fr.loc[ti, "storm_id"]
            ):  # if the priority storm is available that timestep use it
                fr.loc[ti, "storm_id"] = pr_st_noaa
                continue
            # else used method defined
            if method == "first":
                fr.loc[ti, "storm_id"] = fr.loc[ti, "storm_id"][0]
            elif method == "last":
                fr.loc[ti, "storm_id"] = fr.loc[ti, "storm_id"][-1]

    storm_log = np.array(
        [v is not None for v in fr["storm_id"].to_numpy()]
    )  # check for which times there are forecasts from coamps-tc
    if all(~storm_log):  # If no coamps-tc data are available, gfs data are downloaded
        datasets = gfs_forecast_0p25.download(
            param_list, lon_range, lat_range, time_range, cycle_time
        )
        print(
            "Could not find any coamps-tc data in requested time range! Using gfs instead"
        )
        return datasets

    # Else if there are coamps-tc data see which storms (models) to use
    storms_id = np.unique(fr["storm_id"][storm_log])
    dss = {}  # prepare dictionary to save the datasets

    # Loop for the case there is more than one storm in requested times and get storm forecast

    for i, st in enumerate(storms_id):
        dss[st] = {}
        #        storm = storms[[storm_i for storm_i in range(len(storms)) if storms[storm_i]['storm'] == st][0]]
        storm = storms[st]
        t1 = datetime.datetime.strptime(
            storm["min_forecast_date"], "%Y-%m-%d %H:%M:%S"
        ).replace(tzinfo=datetime.timezone.utc)
        t2 = datetime.datetime.strptime(
            storm["max_forecast_date"], "%Y-%m-%d %H:%M:%S"
        ).replace(tzinfo=datetime.timezone.utc)
        # This parts ensures that requested data are within the model range
        t1 = max(requested_times[0], t1)
        t2 = min(requested_times[-1], t2)
        if t1 == t2:  # scrubber tool cannot handle single time
            t1 = t1 - datetime.timedelta(hours=3)
        for ii, var in enumerate(variables):
            ds = met_get(
                domain=[
                    [
                        "coamps-{}".format(st),
                        resolution,
                        lon_range[0],
                        lat_range[0],
                        lon_range[1],
                        lat_range[1],
                    ]
                ],  # list is needed here!
                start=t1.strftime("%Y-%m-%d %H:%M:%S"),
                end=t2.strftime("%Y-%m-%d %H:%M:%S"),
                timestep=timestep,
                variable=var,
                apikey=apikey,
                endpoint=endpoint,
                output="coamps_{}_{}".format(
                    st, var
                ),  # this is used for the online file creation
            )
            if ds is not None:
                dss[st][var] = ds[0]
            else:
                dss[st][var] = None

    # if no data could be downloaded (CHECK AGAIN WHAT SHOULD HAPPEN IN THIS CASE)
    if all(
        [dss[k][kk] is None for k in list(dss.keys()) for kk in list(dss[k].keys())]
    ):
        datasets = gfs_forecast_0p25.download(
            param_list, lon_range, lat_range, time_range, cycle_time
        )
        print(
            "Could not find any coamps-tc data in requested time range! Using gfs instead"
        )
        return datasets
    else:  # Else if there are available data get lon lat info
        for key in dss[storms_id[0]].keys():
            if dss[storms_id[0]][key] is not None:
                data0 = dss[storms_id[0]][key]
                lon = np.array(data0["lon"])
                lat = np.array(data0["lat"])
                if lat[1] - lat[0] > 0:  # lat should be in descending order
                    lat = lat[::-1]
                    reverse = True
                else:
                    reverse = False
                lon = (lon + 360) % 360  # added this to be consistent with gfs
                nrows = len(lat)
                ncols = len(lon)

    # Prepare datasets
    datasets = []
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
        dataset.quantity = param
        dataset.x = lon
        dataset.y = lat
        dataset.time = pd.to_datetime(
            [tt.replace(tzinfo=None) for tt in requested_times]
        ).to_pydatetime()
        if dataset.quantity == "wind":
            dataset.u = np.empty((ntime, nrows, ncols))
            dataset.u[:] = np.NaN
            dataset.v = np.empty((ntime, nrows, ncols))
            dataset.v[:] = np.NaN
        else:
            dataset.val = np.empty((ntime, nrows, ncols))
            dataset.val[:] = np.NaN
        datasets.append(dataset)

    for it, time_i in enumerate(requested_times):
        storm_id = fr["storm_id"][it]

        if storm_id is None:
            try:
                #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
                datasets_gfs = gfs_forecast_0p25.download(
                    param_list, lon_range, lat_range, [time_i, time_i], cycle_time
                )
                for ind, param in enumerate(param_list):
                    if param == "wind":
                        datasets[ind].u[it, :, :] = datasets_gfs[ind].u
                        datasets[ind].v[it, :, :] = datasets_gfs[ind].v
                    else:
                        datasets[ind].val[it, :, :] = datasets_gfs[ind].val
                    datasets[ind].unit = datasets_gfs[ind].unit
                    datasets[ind].src.append("gfs")
                print(
                    "No coamps-tc forecast available for {} --> using gfs forecast instead !!!".format(
                        time_i
                    )
                )

            except Exception:
                print("gfs data not available")
            continue

        # Loop through requested parameters
        for ind, param in enumerate(param_list):
            dataset = datasets[ind]

            try:
                okay = False
                makezeros = False
                #  First check if there are coamps-tc data available for this parameter
                if dss[storm_id][param_names[param]] is not None:
                    okay = True
                    model_t = pd.to_datetime(
                        dss[storm_id][param_names[param]].time.values
                    ).to_pydatetime()
                    model_t_ind = np.where(model_t == time_i.replace(tzinfo=None))[0][0]
                else:
                    makezeros = True

                if okay:
                    dataset.coampsflag = True
                    dataset.src.append("coamps_{}".format(storm_id))
                    if param == "wind":
                        data = dss[storm_id][param_names[param]]
                        u = data["wind_u"]
                        v = data["wind_v"]
                        dataset.unit = u.units

                        if np.any(np.isnan(u)) or np.any(
                            np.isnan(v)
                        ):  # check if there are nans and fill them up
                            u = u.fillna(fill_values[param])
                            v = v.fillna(fill_values[param])

                        u = u.metpy.unit_array.squeeze()
                        v = v.metpy.unit_array.squeeze()

                        if reverse:
                            dataset.u[it, :, :] = np.array(u[model_t_ind, ::-1, :])
                            dataset.v[it, :, :] = np.array(v[model_t_ind, ::-1, :])
                        else:
                            dataset.u[it, :, :] = np.array(u[model_t_ind, :, :])
                            dataset.v[it, :, :] = np.array(v[model_t_ind, :, :])
                    else:
                        if param == "barometric_pressure":
                            var_name = "mslp"
                        elif param == "precipitation":
                            var_name = "rain"  # Check when available in scrubber tool!
                        data = dss[storm_id][param_names[param]]
                        val = data[var_name]

                        # Added this check to ensure that pressure is in Pa
                        if param == "barometric_pressure":
                            if val.units == "mb":
                                val = val * 100
                                val.attrs["units"] = "Pa"
                        dataset.unit = val.units

                        if (
                            param == "precipitation"
                        ):  # Added this check to ensure that nan in precipitation are correctly interpreted
                            val = val.where(val >= 0, np.nan)

                        if np.any(
                            np.isnan(val)
                        ):  # check if there are nans and fill them up
                            val = val.fillna(fill_values[param])

                        val = np.array(val.metpy.unit_array.squeeze())

                        if reverse:
                            dataset.val[it, :, :] = np.array(val[model_t_ind, ::-1, :])
                        else:
                            dataset.val[it, :, :] = np.array(val[model_t_ind, :, :])

                elif makezeros:  # add zeros
                    try:
                        #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
                        datasets_gfs = gfs_forecast_0p25.download(
                            param_list,
                            lon_range,
                            lat_range,
                            [time_i, time_i],
                            cycle_time,
                        )
                        if param == "wind":
                            dataset.u[it, :, :] = datasets_gfs[ind].u
                            dataset.v[it, :, :] = datasets_gfs[ind].v
                        else:
                            dataset.val[it, :, :] = datasets_gfs[ind].val
                        dataset.unit = datasets_gfs[ind].unit
                        dataset.src.append("gfs")
                        print(
                            param
                            + " was not found on server for {} --> using gfs forecast data instead".format(
                                time_i
                            )
                        )
                    except Exception:
                        if param == "wind":
                            dataset.u[:] = fill_values[param]
                            dataset.v[:] = fill_values[param]
                            dataset.unit = units[param]
                            dataset.src.append("default_value")
                            print(
                                param
                                + " was not found on server for {} --> using {} {} instead !!!".format(
                                    time_i, fill_values[param], units[param]
                                )
                            )

                        else:
                            dataset.val[:] = fill_values[param]
                            dataset.unit = units[param]
                            dataset.src.append("default_value")
                            print(
                                param
                                + " was not found on server ... --> using {} {} instead !!!".format(
                                    fill_values[param], units[param]
                                )
                            )

            except Exception:
                print("Could not download data")

    return datasets


def get_priority_storm():
    """Find the storm with the highest priority from NOAA"""
    tcvitals = requests.get("https://ftp.nhc.noaa.gov/atcf/com/tcvitals").text
    splits = [line.split() for line in tcvitals.split("\n")[:-1]]  # last line is empty
    priority = pd.DataFrame(splits)
    priority = priority.iloc[
        [i for i in range(len(priority)) if "L" in priority.iloc[i, 1]], :
    ]  # check only in atlantic basin storms
    try:
        pr_st_noaa = priority.iloc[0, 1]  # define the top L storm as the priority one
        return pr_st_noaa
    except Exception:
        return None


def check_coamps(apikey, endpoint):
    """Read metadata of available forecasts from endpoint and returns a dictionary of the available storms"""
    headers = {"x-api-key": apikey}
    response_API = requests.get("{}/status".format(endpoint), headers=headers)
    data = json.loads(response_API.text)
    storms = data["body"]["data"]["metget"]["coamps-tc"]

    return storms


def parse_domain_data(domain_list: list, level) -> dict:
    import warnings

    AVAILABLE_MODELS = {
        "gfs": "gfs-ncep",
        "nam": "nam-ncep",
        "hwrf": "hwrf",
        "coamps": "coamps-tc",
    }

    model = domain_list[0]
    if "hwrf" in model:
        storm = model.split("-")[1]
        model = "hwrf"
        warnings.warn(
            "HWRF not fully supported yet. Use at your own risk.", RuntimeWarning
        )
    elif "coamps" in model:
        storm = model.split("-")[1]
        model = "coamps"
        # warnings.warn(
        #     "COAMPS not fully supported yet. Use at your own risk.", RuntimeWarning
        # )

    res = float(domain_list[1])
    x0 = float(domain_list[2])
    y0 = float(domain_list[3])
    x1 = float(domain_list[4])
    y1 = float(domain_list[5])

    if model not in AVAILABLE_MODELS.keys():
        raise RuntimeError("Specified model '" + model + "' is not available")

    xmax = max(x0, x1)
    xmin = min(x0, x1)
    ymax = max(y0, y1)
    ymin = min(y0, y1)
    res = abs(res)
    if res <= 0:
        raise RuntimeError("Specified model resolution is invalid")

    if model == "hwrf" or model == "coamps":
        return {
            "name": AVAILABLE_MODELS[model] + "-" + storm,
            "service": AVAILABLE_MODELS[model],
            "storm": storm,
            "x_init": xmin,
            "y_init": ymin,
            "x_end": xmax,
            "y_end": ymax,
            "di": res,
            "dj": res,
            "level": level,
        }
    else:
        return {
            "name": model,
            "service": AVAILABLE_MODELS[model],
            "x_init": xmin,
            "y_init": ymin,
            "x_end": xmax,
            "y_end": ymax,
            "di": res,
            "dj": res,
            "level": level,
        }


def make_metget_request(endpoint, apikey, request_json):
    headers = {"x-api-key": apikey}
    r = requests.post(endpoint + "/build", headers=headers, json=request_json)
    if r.status_code != 200:
        raise RuntimeError(
            "Request to MetGet was returned status code = " + str(r.status_code)
        )
    return_data = json.loads(r.text)
    data_id = return_data["body"]["request_id"]
    status_code = return_data["statusCode"]
    if status_code != 200:
        with open("metget.debug", "a") as f:
            f.write("[WARNING]: MetGet returned status code " + str(status_code) + "\n")
            f.write(str(return_data["body"]["error_text"]))
    return data_id, status_code


def download_metget_data(data_id, endpoint, apikey, sleeptime, max_wait):
    from datetime import datetime, timedelta

    # from netCDF4 import Dataset
    import xarray as xr

    # ...Status check
    headers = {"x-api-key": apikey}
    request_json = {"request": data_id}

    # ...Wait time
    end_time = datetime.utcnow() + timedelta(hours=max_wait)

    # ...Wait for request data to appear
    tries = 0
    data_ready = False
    status = None
    # print("Waiting for request id: ", data_id, flush=True)
    while datetime.utcnow() <= end_time:
        tries += 1
        try:
            # print(
            #     "["
            #     + datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")
            #     + "]: Checking request status...(n="
            #     + str(tries)
            #     + "): ",
            #     flush=True,
            #     end="",
            # )
            response = requests.post(
                endpoint + "/check", headers=headers, json=request_json
            )
            json_response = json.loads(response.text)
            status = json_response["body"]["status"]
            data_url = json_response["body"]["destination"]
            # print(status, flush=True)
            if status == "completed":
                # ...Parse the return to get data
                data_ready = True
                flist_url = data_url + "/filelist.json"
                u = requests.get(flist_url)
                if u.status_code == 200:
                    return_data = json.loads(u.text)
                    # Turned this off to avoid saving this file
                    # with open("filelist.json", "w") as jsonfile:
                    #     jsonfile.write(
                    #         json.dumps(return_data, indent=2, sort_keys=True)
                    #     )
                    break
            elif status == "error":
                # print("Request could not be completed")
                return
            else:
                time.sleep(sleeptime)
                continue
        except KeyboardInterrupt:
            # print("[ERROR]: Process was ended by the user")
            raise

    # ...Download files
    if data_ready:
        file_list = return_data["output_files"]
        ds_list = []
        for f in file_list:
            # print("[{:s}]: Getting file: {:s}".format(time_stamp, f), flush=True)
            url = data_url + "/" + f
            # ds = Dataset('name', memory=requests.get(url).content)
            try:
                ds = xr.open_dataset(
                    url + "#mode=bytes"
                )  # Added this last part to allow opening with xarray
            except Exception:
                from netCDF4 import Dataset

                data = requests.get(url).content
                ds0 = Dataset("temp", memory=data)
                ds = xr.open_dataset(xr.backends.NetCDF4DataStore(ds0))
            ds_list.append(ds)

        return ds_list
    else:
        if status == "restore":
            print(
                "[WARNING]: Data for request "
                + data_id
                + " did not become ready before the max-wait time expired. You can rerun and ask for this request by id"
            )
        elif status == "running":
            print(
                "[WARNING]: Data for request "
                + data_id
                + " is still being constructed when the max-wait time expired. Please check on it later"
            )
        elif status == "queued":
            print(
                "[WARNING]: Data for request "
                + data_id
                + " is still queued. If this does not change soon, please contact an administrator"
            )
        else:
            print("[ERROR]: Data has not become available due to an unknown error")
        return


def met_get(**kwargs):
    """ "based on the get_metget_data.py script but instead of downloading a ntecdf files it returns an xr.dataset
    in memory"""
    import getpass
    import socket
    import types

    AVAILABLE_VARIABLES = {"wind_pressure", "rain", "temperature", "humidity", "ice"}
    AVAILABLE_FORMATS = {"ascii", "owi-ascii", "adcirc-netcdf", "hec-netcdf", "delft3d"}

    args = types.SimpleNamespace()

    for name in kwargs.keys():
        exec('args.{} = kwargs["{}"]'.format(name, name))

    optional = {
        "analysis": False,
        "multiple_forecasts": True,
        "format": "hec-netcdf",
        "variable": "wind_pressure",
        "check_interval": 30,
        "max_wait": 24,
        "strict": True,
        "backfile": True,
        "epsg": 4326,
        "dryrun": False,
        "request": None,
        "compression": False,
        "endpoint": None,
        "apikey": None,
    }

    for key in optional.keys():
        if key not in kwargs:
            exec('args.{} = optional["{}"]'.format(key, key))

    if not args.endpoint:
        if "METGET_ENDPOINT" not in os.environ:
            raise RuntimeError("No endpoint found.")
        else:
            endpoint = os.environ["METGET_ENDPOINT"]
    else:
        endpoint = args.endpoint

    if not args.apikey:
        if "METGET_API_KEY" not in os.environ:
            raise RuntimeError("No API key was found.")
        else:
            apikey = os.environ["METGET_API_KEY"]
    else:
        apikey = args.apikey

    # ...Check for required arguments
    if not args.request:
        if not args.start:
            print("[ERROR]: Must provide '--start'")
            exit(1)
        if not args.end:
            print("[ERROR]: Must provide '--end'")
            exit(1)
        if not args.timestep:
            print("[ERROR]: Must provide '--timestep'")
            exit(1)

        # ...Building the request
        domains = []
        idx = 0
        for d in args.domain:
            j = parse_domain_data(d, idx)
            domains.append(j)
            idx += 1

        if args.format not in AVAILABLE_FORMATS:
            print("ERROR: Invalid output format selected")
            exit(1)

        if args.format == "delft3d" or args.format == "hec-netcdf":
            if len(domains) > 1:
                print(
                    "[ERROR]: "
                    + args.format
                    + " does not support more than one domain."
                )
                exit(1)

        if args.variable not in AVAILABLE_VARIABLES:
            print("ERROR: Invalid variable selected")
            exit(1)

        request_from = getpass.getuser() + "." + socket.gethostname()
        request_data = {
            "version": "0.0.1",
            "creator": request_from,
            "background_pressure": 1013.0,
            "backfill": True,
            "nowcast": args.analysis,
            "multiple_forecasts": args.multiple_forecasts,
            "start_date": str(args.start),
            "end_date": str(args.end),
            "format": args.format,
            "data_type": args.variable,
            "time_step": args.timestep,
            "domains": domains,
            "compression": args.compression,
            "epsg": args.epsg,
            "filename": args.output,
        }
        if args.strict:
            request_data["strict"] = True
        if args.dryrun:
            request_data["dry_run"] = True

        data_id, status_code = make_metget_request(endpoint, apikey, request_data)
        if not args.dryrun and status_code == 200:
            ds_list = download_metget_data(
                data_id, endpoint, apikey, args.check_interval, args.max_wait
            )
        else:
            print(status_code)
            return None

    else:
        ds_list = download_metget_data(
            args.request, endpoint, apikey, args.check_interval, args.max_wait
        )

    return ds_list



# -*- coding: utf-8 -*-
from metget.metget_build import MetGetBuildRest
import tomli
import pandas as pd
import numpy as np
from pyproj import CRS
import datetime
from cht_meteo import gfs_forecast_0p25
from .coamps_utils import *

class Dataset():

    def __init__(self):

        self.quantity = None
        self.unit     = None
        self.time     = []
        self.x        = None
        self.y        = None
        self.crs      = None
        self.val      = None
        self.source   = []
        self.metget   = None


def download(param_list, lon_range, lat_range, time_range, cycle_time, resolution=0.25, config_path=None):
    """Function to download coamps-tc forecasts using the metget tool api
    Right now resolution is hardcoded to 0.25 degrees but this can change"""

    fill_values = dict(wind=0.0, precipitation=0.0, barometric_pressure=102000.0)  # fill values to use in case of nan
    units = dict(wind='m/s', precipitation='kg.m-2.hour-1', barometric_pressure='Pa')  # units

    if not config_path:
        raise ValueError("The path to a configuration file needs to be provided with the argument 'config_path'")
    else:
        with open(config_path, mode="rb") as fp:
            config = tomli.load(fp)
            
    # Read data from the config file
    apikey = config["apikey"]
    endpoint = config["endpoint"]
    api_version = config["api_version"]
    if "tau" in config:
        tau = config["tau"]
    else:
        tau = 0
    if "override" in config:
        override = True
    else:
        override = False

    # Connect parameters names with coamps-tc names
    param_names = {'wind': 'wind_pressure', 'barometric_pressure': 'wind_pressure', 'precipitation': 'rain'}
    variables = list(np.unique([param_names[name] for name in param_list])) # ['wind_pressure', 'rain']

    # Get individual times requested
    # TODO how is time-step handled in cosmos?
    requested_times = pd.date_range(start=time_range[0],
                      end=time_range[1],
                      freq='3H').to_pydatetime().tolist()
    requested_times = [ti.replace(tzinfo=datetime.timezone.utc) for ti in requested_times]

    timestep = int((requested_times[1] - requested_times[0]).total_seconds())  # get time-step of requested time assuming that the time-step stays the same
    ntime = len(requested_times)

    # Get the available storms in the Coamps-TC forecast
    try:
        coamps_storms = check_coamps(apikey, endpoint, time_range[0], time_range[1])
    except:
        coamps_storms = {}
    
    # Get the storm names that have a forecast available for the requested cycle
    storms = [name for name in coamps_storms.keys() if cycle_time.strftime("%Y-%m-%d %H:%M:%S") in coamps_storms[name]["cycles_complete"]]

    # If there are no storms availabe raise Error
    if len(storms) == 0:
        raise ValueError(f"There are no COAMPS-TC forecasts available for cycle: {cycle_time.strftime('%Y-%m-%d %H:%M:%S')}")

    # Check if there is a priority storm given
    if "priority_storm" in config:
        if config["priority_storm"] == "tc_vitals":
            # Get storm priority from noaa
            priority_storm = tc_vitals_storm()
            if not priority_storm:
                print("Priority storm could not be loaded from https://ftp.nhc.noaa.gov/atcf/com/tcvitals")
            else:
                print(f"Priority storm {priority_storm} found from https://ftp.nhc.noaa.gov/atcf/com/tcvitals")
        else:
            priority_storm = config["priority_storm"]
            print(f"Priority storm {priority_storm} provided in config file.")  
    else:
        priority_storm=None
    
    if priority_storm is not None and priority_storm not in storms:
        priority_storm = None
        print(f"Priority storm {priority_storm} not found in available COAMPS-TC forecasted storms {storms}.") 
        datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
        print("Could not find any coamps-tc data in requested time range! Using gfs instead")
        return datasets

    # Choose storm according to input
    if not priority_storm and len(storms) == 1: # if only a single storm is available
        storm = storms[0]
    elif not priority_storm and len(storms) > 1: # if multiple storms are available use latest
        ids = [int(i.split("L")[0]) for i in storms]
        ind = np.argmax(ids)
        storm = storms[ind]
    else:
        storm = priority_storm
    
    print(f"Forecast will be downloaded for storm {storm}")  

    # Prepare metget domain
    domain = [f"coamps-{storm}"] + [resolution] + [lon_range[0], lat_range[0], lon_range[1], lat_range[1]]
    t1 = requested_times[0]
    t2 = requested_times[-1]
    
    dss, metadata = {}, {}
    
    for ii, var in enumerate(variables):
        # ...Building the request
        request_data = MetGetBuildRest.generate_request_json(
            start_date=t1.strftime("%Y%m%d %H%M%S"),
            end_date=t2.strftime("%Y%m%d %H%M%S"),
            format='hec-netcdf',
            data_type=var,
            time_step=timestep,
            domains=MetGetBuildRest.parse_command_line_domains([domain], tau),
            epsg=4326,
            filename=f"{storm}_{var}",
            backfill=True,
            # nowcast=False,
            multiple_forecasts=True, # This makes sure that we use multiple forecasts
            # compression=True,
            save_json_request=False,
            # dry_run=True,
            strict=True,
        )
        client = MetGetBuildRest(endpoint, apikey, api_version)
        data_id, status_code = client.make_metget_request(request_data)
        urls, meta = client.download_metget_data(
            data_id,
            30, # sleep_time in seconds
            3, # max_wait in hours
            output_directory=None,
            return_only_url=True,
        )
        # TODO Check why this should be a list?
        ds_list = []
        for url in urls:
            ds = get_da_from_url(url)
            ds_list.append(ds)
        if ds_list is not None:
            dss[var] = ds_list[0]
            metadata[var] = meta
        else:
            dss[var] = None

    # if no data could be downloaded (CHECK AGAIN WHAT SHOULD HAPPEN IN THIS CASE)
    if all([dss[k] is None for k in list(dss.keys())]):
        datasets = gfs_forecast_0p25.download(param_list, lon_range, lat_range, time_range, cycle_time)
        print("Could not find any coamps-tc data in requested time range! Using gfs instead")
        return datasets
    else: # Else if there are available data get lon lat info
        for key in dss.keys():
            if dss[key] is not None:
                data0 = dss[key]
                lon = np.array(data0['lon'])
                lat = np.array(data0['lat'])
                if lat[1] - lat[0] > 0:  # lat should be in descending order
                    lat = lat[::-1]
                    reverse = True
                else:
                    reverse = False
                lon = (lon + 360) % 360  # added this to be consistent with gfs
                nrows = len(lat)
                ncols = len(lon)

    # Prepare datasets
    datasets = []
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)
        dataset.quantity = param
        dataset.x = lon
        dataset.y = lat
        dataset.time = pd.to_datetime([tt.replace(tzinfo=None) for tt in requested_times]).to_pydatetime()
        if dataset.quantity == "wind":
            dataset.u    = np.empty((ntime, nrows, ncols))
            dataset.u[:] = np.NaN
            dataset.v    = np.empty((ntime, nrows, ncols))
            dataset.v[:] = np.NaN
        else:
            dataset.val    = np.empty((ntime, nrows, ncols))
            dataset.val[:] = np.NaN
        datasets.append(dataset)


    for it, time_i in enumerate(requested_times):
        # Loop through requested parameters
        for ind, param in enumerate(param_list):
            dataset = datasets[ind]
            okay = False
            makezeros = False
            #  First check if there are coamps-tc data available for this parameter
            if dss[param_names[param]] is not None:
                okay = True
                model_t = pd.to_datetime(dss[param_names[param]].time.values).to_pydatetime()
                model_t_ind = np.where(model_t == time_i.replace(tzinfo=None))[0][0]
            else:
                makezeros = True

            if okay:
                # Get metadata of input used
                dataset.coampsflag = True
                dataset.metget = metadata[param_names[param]]["input"]
                inputs = [name for name in metadata[param_names[param]]["input_files"][f"coamps-tc-{storm}"] if time_i.strftime("%Y%m%d%H%M") in name]
                input_name = inputs[0].split(".")[2]
                cycle_used = input_name.split("_")[2]
                cycle_hour = input_name.split("_")[3]
                dataset.source.append(f'coamps_{storm}_{cycle_used}z_{cycle_hour}')
                
                if param == "wind":
                    data = dss[param_names[param]]
                    u = data['wind_u']
                    v = data['wind_v']
                    dataset.unit = u.units
                    
                    if np.any(np.isnan(u)) or np.any(np.isnan(v)):  # check if there are nans and fill them up
                        u = u.fillna(fill_values[param])
                        v = v.fillna(fill_values[param])
                    
                    u = u.metpy.unit_array.squeeze()
                    v = v.metpy.unit_array.squeeze()

                    if reverse:
                        dataset.u[it,:,:] = np.array(u[model_t_ind, ::-1, :])
                        dataset.v[it,:,:] = np.array(v[model_t_ind, ::-1, :])
                    else:
                        dataset.u[it,:,:] = np.array(u[model_t_ind, :, :])
                        dataset.v[it,:,:] = np.array(v[model_t_ind, :, :])
                else:
                    if param == "barometric_pressure":
                        var_name = "mslp"
                    elif param == "precipitation":
                        var_name = 'rain' # Check when available in scrubber tool!
                    data = dss[param_names[param]]
                    val          = data[var_name]

                    # Added this check to ensure that pressure is in Pa
                    if param == "barometric_pressure":
                        if val.units == 'mb':
                            val = val * 100
                            val.attrs['units'] = 'Pa'
                    dataset.unit = val.units

                    if param == "precipitation":  # Added this check to ensure that nan in precipitation are correctly interpreted
                        val = val.where(val>=0, np.nan)

                    if np.any(np.isnan(val)):  # check if there are nans and fill them up
                        val = val.fillna(fill_values[param])
                        
                    val = np.array(val.metpy.unit_array.squeeze())

                    if reverse:
                        dataset.val[it, :, :] = np.array(val[model_t_ind, ::-1, :])
                    else:
                        dataset.val[it, :, :] = np.array(val[model_t_ind, :, :])

            elif makezeros: # add zeros
                try:
                    #  Currently data are used directly from gfs since grid is the same, but  resampling should be done in a new version
                    datasets_gfs = gfs_forecast_0p25.download(param_list, lon_range, lat_range,
                                                            [time_i, time_i], cycle_time)
                    if param == "wind":
                        dataset.u[it, :, :] = datasets_gfs[ind].u
                        dataset.v[it, :, :] = datasets_gfs[ind].v
                    else:
                        dataset.val[it, :, :]  = datasets_gfs[ind].val
                    dataset.unit = datasets_gfs[ind].unit
                    dataset.src.append('gfs')
                    print(param + " was not found on server for {} --> using gfs forecast data instead".format(time_i))
                except:
                    if param == "wind":
                        dataset.u[:] = fill_values[param]
                        dataset.v[:] = fill_values[param]
                        dataset.unit = units[param]
                        dataset.src.append('default_value')
                        print(param + " was not found on server for {} --> using {} {} instead !!!".format(time_i, fill_values[param], units[param]))

                    else:
                        dataset.val[:] = fill_values[param]
                        dataset.unit = units[param]
                        dataset.src.append('default_value')
                        print(param + " was not found on server ... --> using {} {} instead !!!".format(fill_values[param], units[param]))


    return datasets





import os
import pandas as pd
import numpy as np

def write_to_delft3d(
    dataset,
    file_name,
    version="1.03",
    path=None,
    header_comments=False,
    refdate=None,
    parameters=None,
    time_range=None,
):
    
    
    # Convert to datetime
    time = pd.to_datetime(dataset.ds.time.values)

    if not refdate:
        refdate = time[0]

    if not time_range:
        time_range = [time[0], time[-1]]

    if not parameters:
        parameters = []
        for var_name in dataset.ds.data_vars:
            if var_name not in ["lon", "lat", "time"]:
                parameters.append(var_name)

    if dataset.crs.is_geographic:
        grid_unit = "degrees"
    else:
        grid_unit = "m"

    files = []
    for param in parameters:
        # Look up index of this parameter
        for ind, var_name in dataset.ds.data_vars:
            if param == var_name:
                q = var_name
                break
        if param == "wind":
            file = {}
            file["data"] = q.u
            file["ext"] = "amu"
            file["quantity"] = "x_wind"
            file["unit"] = "m s-1"
            file["fmt"] = "%6.1f"
            files.append(file)
            file = {}
            file["data"] = q.v
            file["ext"] = "amv"
            file["quantity"] = "y_wind"
            file["unit"] = "m s-1"
            file["fmt"] = "%6.1f"
            files.append(file)
        elif param == "barometric_pressure":
            file = {}
            file["data"] = q.val
            file["ext"] = "amp"
            file["quantity"] = "air_pressure"
            file["unit"] = "Pa"
            file["fmt"] = "%7.0f"
            files.append(file)
        elif param == "precipitation":
            file = {}
            file["data"] = q.val
            file["ext"] = "ampr"
            file["quantity"] = "precipitation"
            file["unit"] = "mm h-1"
            file["fmt"] = "%7.1f"
            files.append(file)

    # if dataset.quantity == "x_wind":
    #     unit = "m s-1"
    #     ext  = "amu"
    #     fmt  = "%6.1f"
    # elif dataset.quantity == "y_wind":
    #     unit = "m s-1"
    #     ext  = "amv"
    #     fmt  = "%6.1f"
    # elif dataset.quantity == "air_pressure":
    #     unit = "Pa"
    #     ext  = "amp"
    #     fmt  = "%7.0f"
    # elif dataset.quantity == "air_temperature":
    #     unit = "Celsius"
    #     ext  = "amt"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "relative_humidity":
    #     unit = "%"
    #     ext  = "amr"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "cloudiness":
    #     unit = "%"
    #     ext  = "amc"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "sw_radiation_flux":
    #     unit = "W/m2"
    #     ext  = "ams"
    #     fmt  = "%7.1f"
    # elif dataset.quantity == "precipitation":
    #     unit = "mm/h"
    #     ext  = "ampr"
    #     fmt  = "%7.1f"

    for file in files:
        ncols = len(dataset.x)
        nrows = len(dataset.y)

        dx = (dataset.x[-1] - dataset.x[0]) / (len(dataset.x) - 1)
        dy = (dataset.y[-1] - dataset.y[0]) / (len(dataset.y) - 1)

        if path:
            full_file_name = os.path.join(path, file_name + "." + file["ext"])
        else:
            full_file_name = file_name + "." + file["ext"]

        fid = open(full_file_name, "w")

        if header_comments:
            fid.write("### START OF HEADER\n")
            fid.write(
                "### All text on a line behind the first # is parsed as commentary\n"
            )
            fid.write("### Additional comments\n")

        fid.write(
            "FileVersion      =   "
            + version
            + "                                               # Version of meteo input file, to check if the newest file format is used\n"
        )
        fid.write(
            "filetype         =   meteo_on_equidistant_grid                          # Type of meteo input file: meteo_on_flow_grid, meteo_on_equidistant_grid, meteo_on_curvilinear_grid or meteo_on_spiderweb_grid\n"
        )
        fid.write(
            "NODATA_value     =   -999                                               # Value used for undefined or missing data\n"
        )
        fid.write("n_cols           =   " + str(ncols) + "\n")
        fid.write("n_rows           =   " + str(nrows) + "\n")
        fid.write("grid_unit        =   " + grid_unit + "\n")
        #            fid.write("x_llcorner       =   " + str(min(dataset.x)) + "\n")
        #            fid.write("y_llcorner       =   " + str(min(dataset.y)) + "\n")
        fid.write("x_llcorner       =   " + str(min(dataset.x) - 0.5 * dx) + "\n")
        fid.write("y_llcorner       =   " + str(min(dataset.y) - 0.5 * dy) + "\n")
        if version == "1.02":
            fid.write("value_pos       =    corner\n")
        fid.write("dx               =   " + str(dx) + "\n")
        fid.write("dy               =   " + str(dy) + "\n")
        fid.write(
            "n_quantity       =   1                                                  # Number of quantities prescribed in the file\n"
        )
        fid.write("quantity1        =   " + file["quantity"] + "\n")
        fid.write("unit1            =   " + file["unit"] + "\n")
        if header_comments:
            fid.write("### END OF HEADER\n")

        # Add extra blocks if data does not cover time range
        if time[0] > time_range[0]:
            dt = time_range[0] - refdate
            tim = dt.total_seconds() / 60
            val = np.flipud(file["data"][0, :, :])
            # Skip blocks with only nans
            if not np.all(np.isnan(val)):
                val[val == np.nan] = -999.0
                fid.write(
                    "TIME = "
                    + str(tim)
                    + " minutes since "
                    + refdate.strftime("%Y-%m-%d %H:%M:%S")
                    + " +00:00\n"
                )
                np.savetxt(fid, val, fmt=file["fmt"])

        for it, time in enumerate(time):
            dt = time - refdate
            tim = dt.total_seconds() / 60
            val = np.flipud(file["data"][it, :, :])

            if param == "wind":
                if np.max(val) > 1000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    val[np.where(val == 0.0)] = np.nan
                    print(
                        "Warning! Wind speed > 1000 m/s at "
                        + time.strftime("%Y-%m-%d %H:%M:%S")
                        + " !"
                    )
                if np.min(val) < -1000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    print(
                        "Warning! Wind speed > 1000 m/s at "
                        + time.strftime("%Y-%m-%d %H:%M:%S")
                        + " !"
                    )
                    val[np.where(val == 0.0)] = np.nan
            if param == "barometric_pressure":
                if np.max(val) > 200000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    val[np.where(val == 0.0)] = np.nan
                if np.min(val) < 10000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    val[np.where(val == 0.0)] = np.nan
            if param == "precipitation":
                if np.nanmax(val) > 1000.0:
                    val = np.zeros_like(
                        val
                    )  # Weird array, don't trust. Set everything to zeros.
                    print(
                        "Warning! Precipitation exceeds 1000 mm/h at "
                        + time.strftime("%Y-%m-%d %H:%M:%S")
                        + " !"
                    )
                    val[np.where(val == 0.0)] = np.nan
                if np.nanmin(val) < 0.0:
                    val[np.where(val < 0.0)] = 0.0

            if np.all(np.isnan(val)):
                if it > 0:
                    print(
                        "Warning! Only NaNs found for "
                        + param
                        + " at "
                        + time.strftime("%Y-%m-%d %H:%M:%S")
                        + " ! Using data from previous time."
                    )

                else:
                    if param == "wind" or param == "precipitation":
                        print(
                            "Warning! Only NaNs found for "
                            + param
                            + " at "
                            + time.strftime("%Y-%m-%d %H:%M:%S")
                            + " ! Setting values to 0.0 !"
                        )
                        val = np.zeros_like(val)
                    elif param == "barometric_pressure":
                        print(
                            "Warning! Only NaNs found for "
                            + param
                            + " at "
                            + time.strftime("%Y-%m-%d %H:%M:%S")
                            + " ! Setting values to 101300.0 !"
                        )
                        val = np.zeros_like(val) + 101300.0

            fid.write(
                "TIME = "
                + str(tim)
                + " minutes since "
                + refdate.strftime("%Y-%m-%d %H:%M:%S")
                + " +00:00\n"
            )
            np.savetxt(fid, val, fmt=file["fmt"])

        # Add extra blocks if data does not cover time range
        if time[-1] < time_range[1]:
            dt = time_range[1] - refdate
            tim = dt.total_seconds() / 60
            val = np.flipud(file["data"][-1, :, :])
            # Skip blocks with only nans
            if not np.all(np.isnan(val)):
                val[val == np.nan] = -999.0
                fid.write(
                    "TIME = "
                    + str(tim)
                    + " minutes since "
                    + refdate.strftime("%Y-%m-%d %H:%M:%S")
                    + " +00:00\n"
                )
                np.savetxt(fid, val, fmt=file["fmt"])

        fid.close()

def write_wind_to_json(self, file_name, time_range=None, iref=1, js=False):
    import json

    if not time_range:
        time_range = []
        time_range.append(time[0])
        time_range.append(time[-1])

    data = []

    header = {
        "discipline": 0,
        "disciplineName": "Meteorological products",
        "gribEdition": 2,
        "gribLength": 76420,
        "center": 7,
        "centerName": "US National Weather Service - NCEP(WMC)",
        "subcenter": 0,
        "refTime": "2016-04-30T06:00:00.000Z",
        "significanceOfRT": 1,
        "significanceOfRTName": "Start of forecast",
        "productStatus": 0,
        "productStatusName": "Operational products",
        "productType": 1,
        "productTypeName": "Forecast products",
        "productDefinitionTemplate": 0,
        "productDefinitionTemplateName": "Analysis/forecast at horizontal level/layer at a point in time",
        "parameterCategory": 2,
        "parameterCategoryName": "Momentum",
        "parameterNumber": 2,
        "parameterNumberName": "U-component_of_wind",
        "parameterUnit": "m.s-1",
        "genProcessType": 2,
        "genProcessTypeName": "Forecast",
        "forecastTime": 0,
        "surface1Type": 103,
        "surface1TypeName": "Specified height level above ground",
        "surface1Value": 10.0,
        "surface2Type": 255,
        "surface2TypeName": "Missing",
        "surface2Value": 0.0,
        "gridDefinitionTemplate": 0,
        "gridDefinitionTemplateName": "Latitude_Longitude",
        "numberPoints": 65160,
        "shape": 6,
        "shapeName": "Earth spherical with radius of 6,371,229.0 m",
        "gridUnits": "degrees",
        "resolution": 48,
        "winds": "true",
        "scanMode": 0,
        "nx": 360,
        "ny": 181,
        "basicAngle": 0,
        "subDivisions": 0,
        "lo1": 0.0,
        "la1": 90.0,
        "lo2": 359.0,
        "la2": -90.0,
        "dx": 1.0,
        "dy": 1.0,
    }

    header["lo1"] = float(min(dataset.x) + 360.0)
    header["lo2"] = float(max(dataset.x) + 360.0)
    header["la1"] = float(max(dataset.y))
    header["la2"] = float(min(dataset.y))
    header["dx"] = float(dataset.x[1] - dataset.x[0])
    header["dy"] = float(dataset.y[1] - dataset.y[0])
    header["nx"] = len(dataset.x)
    header["ny"] = len(dataset.y)
    header["numberPoints"] = len(dataset.x) * len(dataset.y)

    header_u = header.copy()
    header_v = header.copy()

    header_u["parameterNumberName"] = "U-component_of_wind"
    header_u["parameterNumber"] = 2
    header_v["parameterNumberName"] = "V-component_of_wind"
    header_v["parameterNumber"] = 3

    for it, t in enumerate(time):
        if t >= time_range[0] and t <= time_range[1]:
            dd = []

            tstr = t.strftime("%Y-%m-%dT%H:%M:%SZ")

            u_list = (
                np.flipud(np.around(dataset.quantity[0].u[it, :, :], decimals=1))
                .flatten()
                .tolist()
            )
            data0 = {"header": header_u.copy(), "data": u_list}
            data0["header"]["refTime"] = tstr
            dd.append(data0)

            v_list = (
                np.flipud(np.around(dataset.quantity[0].v[it, :, :], decimals=1))
                .flatten()
                .tolist()
            )
            data0 = {"header": header_v.copy(), "data": v_list}
            data0["header"]["refTime"] = tstr
            dd.append(data0)

            data.append(dd)

    json_string = json.dumps(data, separators=(",", ":"))
    fid = open(file_name, "w")
    if js:
        fid.write("wind = ")
    fid.write(json_string)
    fid.close()



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 10:32:33 2021

@author: ormondt
"""

import numpy as np
import pandas as pd
import xarray as xr
from pyproj import CRS
from siphon.catalog import TDSCatalog
from xarray.backends import NetCDF4DataStore


class Dataset:
    def __init__(self):
        self.quantity = None
        self.unit = None
        self.time = None
        self.x = None
        self.y = None
        self.crs = None
        self.val = None
        self.src = "gfs_forecast_0p25"


def download(param_list, lon_range, lat_range, time_range, cycle_time):
    cycle_string = cycle_time.strftime("%Y%m%d_%H%M")

    base_url = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/"
    url = base_url + "GFS_Global_0p25deg_" + cycle_string + ".grib2/catalog.xml"
    url = "https://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/catalog.xml?dataset=grib/NCEP/GFS/Global_0p25deg/Best"

    gfs = TDSCatalog(url)
    ds = list(gfs.datasets.values())[0]
    ncss = ds.subset()

    datasets = []

    # Loop through requested parameters
    for param in param_list:
        dataset = Dataset()
        dataset.crs = CRS.from_epsg(4326)

        if param == "wind":
            dataset.quantity = param

            query = ncss.query()
            query.lonlat_box(
                north=lat_range[1],
                south=lat_range[0],
                east=lon_range[1],
                west=lon_range[0],
            ).time_range(time_range[0], time_range[1]).vertical_level(10.0)
            query.variables(
                "u-component_of_wind_height_above_ground",
                "v-component_of_wind_height_above_ground",
            )
            data = ncss.get_data(query)
            data = xr.open_dataset(NetCDF4DataStore(data))
            try:
                dataset.x = np.array(data["lon"])
                dataset.y = np.array(data["lat"])
            except Exception:
                dataset.x = np.array(data["longitude"])
                dataset.y = np.array(data["latitude"])
            #           time   = data['time']

            u = data["u-component_of_wind_height_above_ground"]
            time = find_time_var(u)
            dataset.time = pd.to_datetime(time.data).to_pydatetime()
            dataset.unit = u.units
            u = u.metpy.unit_array.squeeze()
            dataset.u = np.array(u)
            v = data["v-component_of_wind_height_above_ground"]
            v = v.metpy.unit_array.squeeze()
            dataset.v = np.array(v)

        else:
            # Other scalar variables

            fac = 1.0

            dataset.quantity = param

            if param == "barometric_pressure":
                var_name = "Pressure_reduced_to_MSL_msl"
            elif param == "precipitation":
                var_name = "Precipitation_rate_surface"
                fac = 3600.0

            query = ncss.query()
            query.lonlat_box(
                north=lat_range[1],
                south=lat_range[0],
                east=lon_range[1],
                west=lon_range[0],
            ).time_range(time_range[0], time_range[1])
            query.variables(var_name)
            data = ncss.get_data(query)
            data = xr.open_dataset(NetCDF4DataStore(data))

            try:
                dataset.x = np.array(data["lon"])
                dataset.y = np.array(data["lat"])
            except Exception:
                dataset.x = np.array(data["longitude"])
                dataset.y = np.array(data["latitude"])
            #            time   = data['time']
            #            d.time = pd.to_datetime(time.data).to_pydatetime()

            val = data[var_name]
            time = find_time_var(val)
            dataset.time = pd.to_datetime(time.data).to_pydatetime()
            dataset.unit = val.units
            val = val.metpy.unit_array.squeeze()
            dataset.val = np.array(val) * fac

        datasets.append(dataset)

    return datasets


# Helper function for finding proper time variable
def find_time_var(var, time_basename="time"):
    for coord_name in var.coords:
        if coord_name.startswith(time_basename):
            return var.coords[coord_name]
    raise ValueError("No time variable found for " + var.name)


# def write_wind_to_nc(lon, lat, time, u, v, meteo_name, meteo_path):

#     for it, t in enumerate(time):

#         time_string = t.strftime("%Y%m%d_%H%M")
#         file_name = meteo_name + ".wind." + time_string + ".f.nc"
#         full_file_name = os.path.join(meteo_path, file_name)
#         uu = u[it,:,:]
#         vv = v[it,:,:]
#         ds = xr.Dataset({"u": (("lat", "lon"), uu),
#                          "v": (("lat", "lon"), vv)},
#                         coords={
#                         "lon": lon,
#                         "lat": lat})
#         ds.to_netcdf(path=full_file_name)

# def write_p_to_nc(lon, lat, time, p, meteo_name, meteo_path):

#     for it, t in enumerate(time):

#         time_string = t.strftime("%Y%m%d_%H%M")
#         file_name = meteo_name + ".p." + time_string + ".f.nc"
#         full_file_name = os.path.join(meteo_path, file_name)
#         pp = p[it,:,:]
#         ds = xr.Dataset({"p": (("lat", "lon"), pp)},
#                         coords={
#                         "lon": lon,
#                         "lat": lat})
#         ds.to_netcdf(path=full_file_name)

# def write_pr_to_nc(lon, lat, time, pr, meteo_name, meteo_path):

#     for it, t in enumerate(time):

#         time_string = t.strftime("%Y%m%d_%H%M")
#         file_name = meteo_name + ".precip." + time_string + ".f.nc"
#         full_file_name = os.path.join(meteo_path, file_name)
#         pp = pr[it,:,:]
#         ds = xr.Dataset({"precip": (("lat", "lon"), pp)},
#                         coords={
#                         "lon": lon,
#                         "lat": lat})
#         ds.to_netcdf(path=full_file_name)
